/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_CORE_H_
#define _CTRL_CORE_H_

#ifndef CTRL_FPGA_KERNEL_FILE
#include <omp.h>
#endif

#ifdef _CTRL_DEBUG_
    #include <stdio.h>
#endif

#ifndef CTRL_FPGA_KERNEL_FILE
#include "hitmap2.h"
#else
#include "hit_kernel.h"
#endif

#ifndef CTRL_FPGA_KERNEL_FILE
#include "Core/Ctrl_Policy.h"
#include "Core/Ctrl_TaskQueue.h"
#include "Core/Ctrl_Type.h"

#include "Kernel/Ctrl_KernelArgs.h"
#endif

#ifdef _CTRL_ARCH_CPU_
	#include "Architectures/Cpu/Ctrl_Cpu.h"
#else
	#define CTRL_CPU_LAUNCH( ... )
#endif // _CTRL_ARCH_CPU_


#ifdef _CTRL_ARCH_CUDA_LEGACY_
	#include "Architectures/Cuda/Legacy/Ctrl_Cuda.h"
#elif _CTRL_ARCH_CUDA_
	#include "Architectures/Cuda/Ctrl_Cuda.h"
#else
	#define CTRL_CUDA_LAUNCH( ... )
#endif // _CTRL_ARCH_CUDA_


#ifdef _CTRL_ARCH_OPENCL_GPU_
	#include "Architectures/OpenCL/Ctrl_OpenCL_Gpu.h"
#else
	#define CTRL_OPENCL_GPU_LAUNCH( ... )
#endif // _CTRL_ARCH_OPENCL_GPU_


#ifdef _CTRL_ARCH_FPGA_
	#ifndef CTRL_FPGA_KERNEL_FILE
		#include "Architectures/FPGA/Ctrl_FPGA.h"
	#endif
#else
	#define CTRL_FPGA_LAUNCH( ... )
#endif // _CTRL_ARCH_FPGA_


#ifdef _CTRL_ARCH_XPHI_
	#include "Architectures/XPhi/Ctrl_XPhi.h"
#else 
	#define CTRL_XPHI_LAUNCH( ... )
#endif // _CTRL_ARCH_XPHI_


#ifdef __cplusplus
	extern "C" {
#endif




#ifndef CTRL_FPGA_KERNEL_FILE
#define MAX_BLOCK_SIZE	256
/**
 * @brief Contiene los datos de las estructuras.
 * Se utiliza una unión por eficiencia.
 */
typedef union {

	#ifdef _CTRL_ARCH_CPU_
		struct Ctrl_Cpu cpu;
	#endif // _CTRL_ARCH_CPU_

	#ifdef _CTRL_ARCH_CUDA_
		struct Ctrl_Cuda cuda; 
	#endif // _CTRL_ARCH_CUDA_

	#ifdef _CTRL_ARCH_OPENCL_GPU_
		struct Ctrl_OpenCLGpu opencl_gpu;
	#endif // _CTRL_ARCH_OPENCL_GPU_

	#ifdef _CTRL_ARCH_FPGA_
		struct Ctrl_FPGA fpga;
	#endif // _CTRL_ARCH_OPENCL_GPU_

	#ifdef _CTRL_ARCH_XPHI_
		struct Ctrl_XPhi xphi;
	#endif // _CTRL_ARCH_XPHI_

} Ctrl_Impl;


/**
 * @brief Controlador abstracto.
 */
typedef struct {
	int 				id;
	Ctrl_Type			type;
	Ctrl_Impl			*p_impl;
	
	// sem_t				sem_destroy;
	// sem_t				sem_sync;

	// omp_lock_t			lock_queue;
	// omp_lock_t			lock_sync;

	omp_lock_t			*p_lock_first_host;
	omp_lock_t			*p_lock_first_ctrl;
	omp_lock_t			*p_lock_host;
	omp_lock_t			*p_lock_ctrl;
	// omp_lock_t			lock_r2;
	// omp_lock_t			lock_w2;

	// bool				syncronized;

	Ctrl_TaskQueue		*p_task_queue;
	// int					n_task;
	// int					n_innerTask;

	Ctrl_Policy			policy;
	int					device;
	void				*p_ext;
	
} Ctrl;

typedef Ctrl * PCtrl;

Ctrl Ctrl_global;


#define Ctrl_Launch( p_ctrl, name, threads, group, ... ) \
	switch ((p_ctrl)->type) \
	{ \
		CTRL_CPU_LAUNCH( p_ctrl, name, threads, group, __VA_ARGS__ ) \
		CTRL_CUDA_LAUNCH( p_ctrl, name, threads, group, __VA_ARGS__ ) \
		CTRL_OPENCL_GPU_LAUNCH( p_ctrl, name, threads, group, __VA_ARGS__ ) \
		CTRL_FPGA_LAUNCH( p_ctrl, name, threads, group, __VA_ARGS__ ) \
		CTRL_XPHI_LAUNCH( p_ctrl, name, threads, group, __VA_ARGS__ ) \
	}

#define Ctrl_HostTask( p_ctrl, name, ... ) \
	Ctrl_LaunchHostTask(p_ctrl, Ctrl_HostTaskCreate_##name( CTRL_KERNEL_ARGS_TO_POINTERS( __VA_ARGS__) )); \

// #define __ctrl_block__( num_ctrl ) omp_set_nested( 1 ); omp_set_num_threads( 1 + (num_ctrl) ); _Pragma("omp parallel") _Pragma("omp master")
// #define __ctrl_block__( num_ctrl ) omp_set_nested( 1 ); omp_set_num_threads( 1 + 2 * (num_ctrl) ); _Pragma("omp parallel") _Pragma("omp master") 
// #define __ctrl_block__( num_ctrl ) omp_set_nested( 1 ); _Pragma("omp parallel") _Pragma("omp single")
// extern int Ctrl_n_ctrls;
// extern Ctrl *Ctrl_ctrls[num_ctrl];
// extern omp_lock_t Ctrl_locks[4 * num_ctrl];

// #define __ctrl_block__(num_ctrl)                               \
// 	int Ctrl_n_ctrls = num_ctrl;                               \
// 	Ctrl *Ctrl_ctrls[num_ctrl];                                \
// 	omp_lock_t Ctrl_locks[4 * num_ctrl];                       \
// 	for (int i = 0; i < num_ctrl) {                            \
// 		Ctrl_ctrls[i] = (Ctrl *)malloc(sizeof(Ctrl));          \
// 		Ctrl_ctrls[i]->id = i;                                 \
// 		\ Ctrl_TaskQueue_Init(&((Ctrl_ctrls[i])->task_queue)); \
// 		for (int j = 0; j < 4; j++) {                          \
// 			omp_init_lock(&(Ctrl_locks[i * 4 + j]));           \
// 		}                                                      \
// 	}                                                          \
// 	omp_set_nested(1);                                         \
// 	omp_set_num_threads(1 + 2 * (num_ctrl));                   \
// 	_Pragma("omp parallel") if (Ctrl_Thread_init() == 0)



#define __ctrl_block__(num_ctrl)                               \
	omp_set_nested(1);                                         \
	omp_set_num_threads(1 + 2 * (num_ctrl));                   \
	_Pragma("omp parallel") if (Ctrl_Thread_init() == 0)

// #ifdef _CTRL_ARCH_CUDA_
// 	#define __ctrl_block__( num_ctrl ) \
// 		omp_set_nested( 1 ); \
// 		omp_set_num_threads( 1 + 1 * (num_ctrl) ); \
// 		_Pragma("omp parallel") \
// 		if (Ctrl_Thread_init() == 0)
// #else
// 	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
// 		#define __ctrl_block__( num_ctrl ) \
// 			omp_set_nested( 1 ); \
// 			omp_set_num_threads( 1 + 2 * (num_ctrl) ); \
// 			_Pragma("omp parallel") \
// 			if (Ctrl_Thread_init() == 0)
// 	#else
// 		#define __ctrl_block__( num_ctrl ) \
// 			omp_set_nested( 1 ); \
// 			omp_set_num_threads( 1 + 1 * (num_ctrl) ); \
// 			_Pragma("omp parallel") \
// 			if (Ctrl_Thread_init() == 0)
// 	#endif
// #endif


/* @author: Gabriel Rodriguez-Canal
   @brief: Kernel files need KHitTile definitions to be able to recover them from wrappers */
#define Ctrl_NewType( type ) \
		\
		hitNewType( type ) ; \
		hit_ktileNewType( type ); \
		static inline HitTile_##type Ctrl_Alloc_##type( Ctrl *p_ctrl, HitShape shape) { \
			HitTile_##type new_tile = HIT_TILE_NULL_STATIC; \
			hit_tileDomainShape( &new_tile, type, shape ); \
			Ctrl_AllocInner( p_ctrl, ((HitTile *)(&new_tile)) ); \
			return new_tile; \
		} \
		static inline HitTile_##type Ctrl_AllocInternal_##type( Ctrl *p_ctrl, HitShape shape) { \
			HitTile_##type new_tile = HIT_TILE_NULL_STATIC; \
			hit_tileDomainShape( &new_tile, type, shape ); \
			Ctrl_AllocInternalInner( p_ctrl, ((HitTile *)(&new_tile)) ); \
			return new_tile; \
		} \
		\
		typedef struct { \
			int origAcumCard[4]; \
			int card[3]; \
		} fpga_wrapper_KHitTile_##type;
#else
#ifndef CTRL_HOST_COMPILE
	#ifndef CTRL_FPGA_BIN_NAME
		#define Ctrl_NewType( type ) \
				hit_ktileNewType( type ); \
				\
				typedef struct { \
					int origAcumCard[4]; \
					int card[3]; \
				} fpga_wrapper_KHitTile_##type; \
				\
				typedef __global void* restrict data_KHitTile_##type;
	#else
		#define Ctrl_NewType( type )
	#endif
#else
	#define Ctrl_NewType( type )
#endif
#endif

#ifndef CTRL_FPGA_KERNEL_FILE

/*
	@author: Gabriel Rodriguez-Canal
	Wrapper for controller creation. ## operator before __VA_ARGS__ removes trailing comma when only policy and device are passed (CUDA).
*/
#define Ctrl_Create( type, policy, device, ... ) Ctrl_CreateWrapper_##type(policy, device, ##__VA_ARGS__ );

#ifdef _CTRL_ARCH_FPGA_
	PCtrl Ctrl_CreateWrapper_CTRL_TYPE_FPGA(Ctrl_Policy policy, int device, int platform, int exec_mode);
#endif //TODO: remaining architectures

#ifdef _CTRL_ARCH_OPENCL_GPU_
	PCtrl Ctrl_CreateWrapper_CTRL_TYPE_OPENCL_GPU(Ctrl_Policy policy, int device, int platform);
#endif

#ifdef _CTRL_ARCH_CUDA_
	PCtrl Ctrl_CreateWrapper_CTRL_TYPE_CUDA(Ctrl_Policy policy, int device);
#endif



#define Ctrl_Alloc(ctrl, type, shape) Ctrl_Alloc_##type( ctrl, shape );
#define Ctrl_AllocInternal(ctrl, type, shape) Ctrl_AllocInternal_##type( ctrl, shape );

void Ctrl_AllocInner(Ctrl *p_ctrl, HitTile *p_tile);
void Ctrl_AllocInternalInner(Ctrl *p_ctrl, HitTile *p_tile);

#define Ctrl_Free( ctrl, ... ) Ctrl_FreeN( ctrl, CTRL_COUNTPARAM( __VA_ARGS__ ), __VA_ARGS__ )
#define Ctrl_FreeN( ctrl, n_tiles, ... ) Ctrl_FreeN2( ctrl, n_tiles, __VA_ARGS__ )
#define Ctrl_FreeN2( ctrl, n_tiles, ... ) Ctrl_OP_##n_tiles( Free, ctrl, __VA_ARGS__ )
void Ctrl_FreeInner(Ctrl *p_ctrl, HitTile *p_tile);

#define Ctrl_MoveTo( ctrl, ... ) Ctrl_MoveToN( ctrl, CTRL_COUNTPARAM( __VA_ARGS__ ), __VA_ARGS__ )
#define Ctrl_MoveToN( ctrl, n_tiles, ... ) Ctrl_MoveToN2( ctrl, n_tiles, __VA_ARGS__ )
#define Ctrl_MoveToN2( ctrl, n_tiles, ... ) Ctrl_OP_##n_tiles( MoveTo, ctrl, __VA_ARGS__ )
void Ctrl_MoveToInner(Ctrl *p_ctrl, HitTile *p_tile);

#define Ctrl_MoveFrom( ctrl, ... ) Ctrl_MoveFromN( ctrl, CTRL_COUNTPARAM( __VA_ARGS__ ), __VA_ARGS__ )
#define Ctrl_MoveFromN( ctrl, n_tiles, ... ) Ctrl_MoveFromN2( ctrl, n_tiles, __VA_ARGS__ )
#define Ctrl_MoveFromN2( ctrl, n_tiles, ... ) Ctrl_OP_##n_tiles( MoveFrom, ctrl, __VA_ARGS__ )
void Ctrl_MoveFromInner(Ctrl *p_ctrl, HitTile *p_tile);

#define Ctrl_WaitTile( ctrl, ... ) Ctrl_WaitTileN( ctrl, CTRL_COUNTPARAM( __VA_ARGS__ ), __VA_ARGS__ )
#define Ctrl_WaitTileN( ctrl, n_tiles, ... ) Ctrl_WaitTileN2( ctrl, n_tiles, __VA_ARGS__ )
#define Ctrl_WaitTileN2( ctrl, n_tiles, ... ) Ctrl_OP_##n_tiles( WaitTile, ctrl, __VA_ARGS__ )
void Ctrl_WaitTileInner(Ctrl *p_ctrl, HitTile *p_tile);

void Ctrl_GlobalSync(Ctrl *p_ctrl);

void Ctrl_LaunchKernel(Ctrl *p_ctrl, Ctrl_Task task);

void Ctrl_LaunchHostTask(Ctrl *p_ctrl, Ctrl_Task task);

int Ctrl_Thread_init();

void Ctrl_Destroy(Ctrl *p_ctrl);

int Ctrl_Dev(Ctrl_Type type, int sum_impl);

#define Ctrl_OP_1( op, ctrl, tile )   Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile)));
#define Ctrl_OP_2( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_1( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_3( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_2( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_4( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_3( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_5( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_4( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_6( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_5( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_7( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_6( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_8( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_7( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_9( op, ctrl, tile, ... )  Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_8( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_10( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_9( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_11( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_10( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_12( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_11( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_13( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_12( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_14( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_13( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_15( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_14( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_16( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_15( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_17( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_16( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_18( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_17( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_19( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_18( op, ctrl, __VA_ARGS__ )
#define Ctrl_OP_20( op, ctrl, tile, ... ) Ctrl_##op##Inner(ctrl, ((HitTile *)&(tile))); Ctrl_OP_19( op, ctrl, __VA_ARGS__ )

#ifdef __cplusplus
	}
#endif

#endif // CTRL_FPGA_KERNEL_FILE

#endif /* _CTRL_CORE_H_ */
