/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
/*
 * CTRL: Task queues inside the controllers
 *
 * @version 2.0
 * @author Arturo Gonzalez-Escrinano
 * @date Sep 2016
 */

#ifndef _CTRL_CORE_TASKQUEUE_H_
#define _CTRL_CORE_TASKQUEUE_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>

#include "Core/Ctrl_Request.h"

#include "Kernel/Ctrl_ImplType.h"
#include "Kernel/Ctrl_Thread.h"

typedef enum {
    CTRL_TASK_TYPE_NULL,
    CTRL_TASK_TYPE_KERNEL,
    CTRL_TASK_TYPE_HOST,
    CTRL_TASK_TYPE_DESTROYCNTRL,
    CTRL_TASK_TYPE_GLOBALSYNC,
    CTRL_TASK_TYPE_ALLOCTILE,
    CTRL_TASK_TYPE_FREETILE,
    CTRL_TASK_TYPE_MOVETO,
    CTRL_TASK_TYPE_MOVEFROM,
    CTRL_TASK_TYPE_WAITTILE,
} Ctrl_TaskType;

#define CTRL_TASK_NAME_MAX_LEN 29

typedef struct Ctrl_Task {
    int device_id; // The device id inside the Ctrl
    Ctrl_TaskType task_type; // A task label (Future optimization: Reuse predefined tasks)
    void (*pfn_kernel_wrapper)
			(
				Ctrl_Request request, Ctrl_ImplType impl_type, int impl,
				Ctrl_Thread threads,
				void *p_arguments
			); // Kernel launching wrapper pointer
    void (*pfn_hostTask_wrapper)(void *p_arguments); // Host Task launching wrapper pointer
    int n_arguments;        // Number of arguments/roles/pointers
    void *p_arguments;      // Packed list of arguments
    char *p_roles;          // Input/Output roles, for memory optimizations
    void **pp_pointers; // Pointers to the original variables, for memory basic operations
    uint8_t *p_displacements; // Displacement of parameter over arguments array, for memory basic operations
    struct Ctrl_Task *p_next; // Next task in the queue
    Ctrl_Thread threads;      // Index domain where the task is executed
    HitTile *p_tile;        // For 1 tile tasks
} Ctrl_Task;


typedef struct {
    int read;
    int write;
    Ctrl_Task buffer[50000];
} Ctrl_TaskQueue;

#define CTRL_DEVICE_ALL -1

#define CTRL_TASK_NULL                                                \
    {                                                                 \
        .device_id = 0, .task_type = CTRL_TASK_TYPE_NULL,             \
        .pfn_kernel_wrapper = NULL, .pfn_hostTask_wrapper = NULL,     \
        .n_arguments = 0, .p_arguments = NULL, .p_roles = NULL,       \
        .pp_pointers = NULL, .p_displacements = NULL, .p_next = NULL, \
        .threads = CTRL_THREAD_NULL, .p_tile = NULL                    \
    }

static inline void Ctrl_TaskQueue_FreeTask(Ctrl_Task *p_task) {
    p_task->device_id = 0;
    p_task->pfn_kernel_wrapper = NULL;
    p_task->task_type = CTRL_TASK_TYPE_NULL;
    p_task->pfn_hostTask_wrapper = NULL;
    p_task->n_arguments = 0;
    if ((p_task->p_arguments) != NULL) free(p_task->p_arguments);
    if ((p_task->p_roles) != NULL) free(p_task->p_roles);
    if ((p_task->pp_pointers) != NULL) free(p_task->pp_pointers);
    if ((p_task->p_displacements) != NULL) free(p_task->p_displacements);
    p_task->p_next = NULL;
    p_task->p_tile = NULL;
}

static inline void Ctrl_TaskQueue_Init(Ctrl_TaskQueue *p_queue) {
    p_queue->read = p_queue->write = 0;
}

static inline void Ctrl_TaskQueue_Push(Ctrl_TaskQueue *p_queue, Ctrl_Task task) {
    p_queue->buffer[p_queue->write] = task;

    #pragma omp atomic update
    p_queue->write++;
}

static inline Ctrl_Task *Ctrl_TaskQueue_Pop(Ctrl_TaskQueue *p_queue) {
    int write = 0;

    #pragma omp atomic read
    write = p_queue->write;

    while(write == p_queue->read) {
        #pragma omp atomic read
        write = p_queue->write;
    }

    p_queue->read++;
    return &(p_queue->buffer[(p_queue->read) - 1]);
}

static inline void Ctrl_TaskQueue_Destroy(Ctrl_TaskQueue *p_queue) {
    for (int i = 0; i < p_queue->read; i++){
        if ((p_queue->buffer[i]).task_type == CTRL_TASK_TYPE_HOST) {
            Ctrl_TaskQueue_FreeTask(&(p_queue->buffer[i]));
        }
    }
    p_queue->read = p_queue->write = 0;
}

#ifdef _CTRL_DEBUG_
    static inline void Ctrl_TaskQueue_GetTypeName(Ctrl_Task *p_task, char *name) {
        switch(p_task->task_type){
            case CTRL_TASK_TYPE_NULL:
                strcpy(name, "CTRL_TASK_TYPE_NULL");
                break;
            case CTRL_TASK_TYPE_KERNEL:
                strcpy(name, "CTRL_TASK_TYPE_KERNEL");
                break;
            case CTRL_TASK_TYPE_HOST:
                strcpy(name, "CTRL_TASK_TYPE_HOST");
                break;
            case CTRL_TASK_TYPE_DESTROYCNTRL:
                strcpy(name, "CTRL_TASK_TYPE_DESTROYCNTRL");
                break;
            case CTRL_TASK_TYPE_GLOBALSYNC:
                strcpy(name, "CTRL_TASK_TYPE_GLOBALSYNC");
                break;
            case CTRL_TASK_TYPE_ALLOCTILE:
                strcpy(name, "CTRL_TASK_TYPE_ALLOCTILE");
                break;
            case CTRL_TASK_TYPE_FREETILE:
                strcpy(name, "CTRL_TASK_TYPE_FREETILE");
                break;
            case CTRL_TASK_TYPE_MOVETO:
                strcpy(name, "CTRL_TASK_TYPE_MOVETO");
                break;
            case CTRL_TASK_TYPE_MOVEFROM:
                strcpy(name, "CTRL_TASK_TYPE_MOVEFROM");
                break;
            case CTRL_TASK_TYPE_WAITTILE:
                strcpy(name, "CTRL_TASK_TYPE_WAITTILE");
                break;
            case CTRL_TASK_TYPE_ALLOCINTERNAL:
                strcpy(name, "CTRL_TASK_TYPE_ALLOCINTERNAL");
                break;
            default:
                strcpy(name, "Unknown task type");
                break;
        }
    }
#endif // _CTRL_DEBUG_

#endif // _CTRL_CORE_TASKQUEUE_H_
