/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include <stdbool.h>
#include <stdio.h>

#include "Core/Ctrl_Request.h"
#include "Core/Ctrl_Core.h"

#include "Architectures/Cpu/Ctrl_Cpu.h"
#include "Architectures/Cpu/Ctrl_Cpu_Request.h"



/*************************************************************
 ******** Prototypes of private functions ********************
 *************************************************************/

void Ctrl_Cpu_ConsumeTaskQueue(Ctrl_Cpu *p_ctrl);

/*************************************************************
 ******** Prototypes of tasks' evaluation functions **********
 *************************************************************/
void Ctrl_Cpu_EvalTaskKernelLaunch(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cpu_EvalTaskGlobalSync(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cpu_EvalTaskAttachTile(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cpu_EvalTaskDetachTile(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cpu_EvalTaskMoveTo(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cpu_EvalTaskMoveFrom(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cpu_EvalTaskWaitTile(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task);

/********************************************
 ******** GPU Controller functions **********
 ********************************************/

/*
 * Create the controller and its corresponding variables
 * Params:
 *      cntrl: Controller to be created.
 *      device: GPU used by this controller
 */
void Ctrl_Cpu_Create(Ctrl_Cpu *p_ctrl, int n_cores, Ctrl_Policy policy) {
	
	if(policy != CTRL_POLICY_DEFAULT) {
		p_ctrl->policy = policy;
	} else {
		if ( getenv("CTRL_ASYNC_MODE") ) {
			if (atoi( getenv("CTRL_ASYNC_MODE") ) > 0) {
				p_ctrl->policy = CTRL_POLICY_ASYNC;
			} else {
				p_ctrl->policy = CTRL_POLICY_SYNC;
			}
		} else {
			p_ctrl->policy = CTRL_POLICY_SYNC;
		}
	}

	p_ctrl->n_task = 0;

	/* Allocate the controller */
	Ctrl_TaskQueue_Init(&(p_ctrl->task_queue));


	/* Initialize the data structure */
	p_ctrl->n_cores = n_cores;

	// omp_init_lock(&(p_ctrl->destroy_ctrl));
	// omp_init_lock(&(p_ctrl->global_sync));
	omp_init_lock(&(p_ctrl->lock_queue));
	
	sem_init(&(p_ctrl->sem_destroy), 0, 1);
	sem_init(&(p_ctrl->sem_sync), 0, 1);

	sem_wait(&(p_ctrl->sem_destroy));

	//TODO @arturo afinidades
	#pragma omp task
	{
		{
			{
				Ctrl_Cpu_ConsumeTaskQueue(p_ctrl);
				
				omp_set_lock(&(p_ctrl->lock_queue));
				Ctrl_TaskQueue_Clean(&(p_ctrl->task_queue));
				omp_unset_lock(&(p_ctrl->lock_queue));

				omp_destroy_lock(&(p_ctrl->lock_queue));

				sem_post(&(p_ctrl->sem_destroy));
			}
		}
	}
}


/*
 * Synchornous policy to consume queue of tasks 
 * Params:
 *      ctrl: instance of Controller.
 */
void Ctrl_Cpu_ConsumeTaskQueue(Ctrl_Cpu *p_ctrl) {
	
	int task_done = 0;
	bool finish = false;

	while (!finish) {

		omp_set_lock(&(p_ctrl->lock_queue));
		if ( !(Ctrl_TaskQueue_IsEmpty(&(p_ctrl->task_queue))) ) {
			
			Ctrl_Task *p_task = Ctrl_TaskQueue_Pop( &(p_ctrl->task_queue) );
			omp_unset_lock(&(p_ctrl->lock_queue));

			// Evaluation of task
			switch ( p_task->task_type ) {
				case CTRL_TASK_TYPE_KERNEL: 
					Ctrl_Cpu_EvalTaskKernelLaunch(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_DESTROYCNTRL:
					finish = true;
					break;
				case CTRL_TASK_TYPE_GLOBALSYNC:
					Ctrl_Cpu_EvalTaskGlobalSync(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_ATTACH:
					Ctrl_Cpu_EvalTaskAttachTile(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_DETACH: 
					Ctrl_Cpu_EvalTaskDetachTile(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_MOVETO:
					Ctrl_Cpu_EvalTaskMoveTo(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_MOVEFROM:
					Ctrl_Cpu_EvalTaskMoveFrom(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_WAITTILE:
					Ctrl_Cpu_EvalTaskWaitTile(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_CREATEINTERNAL:
					Ctrl_Cpu_EvalTaskAttachTile(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_DESTROYINTERNAL:
					Ctrl_Cpu_EvalTaskDetachTile(p_ctrl, p_task);
					break;
			}
			task_done++;
			free(p_task);
		} else {
			omp_unset_lock(&(p_ctrl->lock_queue));
		}
	}
}


/*********************************
 * CONTROLLER EXTERNAL FUNCTIONS *
 *********************************/

//Add task to the cntrl
void Ctrl_Cpu_AddTaskToQueue(Ctrl_Cpu *p_ctrl, Ctrl_Task task) {
	task.task_id = p_ctrl->n_task;

	omp_set_lock(&(p_ctrl->lock_queue));
	Ctrl_TaskQueue_Push(&(p_ctrl->task_queue), task);
	omp_unset_lock(&(p_ctrl->lock_queue));

	p_ctrl->n_task += 1;
}

void Ctrl_Cpu_LaunchKernel(Ctrl_Cpu *p_ctrl, Ctrl_Task task) {
	Ctrl_Cpu_AddTaskToQueue(p_ctrl, task);
}


/*
 * Destroy a Controller
 */
void Ctrl_Cpu_Destroy(Ctrl_Cpu *p_ctrl) {
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_DESTROYCNTRL;
	
	Ctrl_Cpu_AddTaskToQueue(p_ctrl, task);

	sem_wait(&(p_ctrl->sem_destroy));
	sem_destroy(&(p_ctrl->sem_destroy));
	sem_destroy(&(p_ctrl->sem_sync));
}


void Ctrl_Cpu_GlobalSync(Ctrl_Cpu *p_ctrl){
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_GLOBALSYNC;

	Ctrl_Cpu_AddTaskToQueue(p_ctrl, task);
	
	sem_wait(&(p_ctrl->sem_sync));
}


/*
 * Attach a tile
 */
void Ctrl_Cpu_Attach(Ctrl_Cpu *p_ctrl, HitTile *p_tile) {}


/*
 * Move a tile to device
 */
void Ctrl_Cpu_MoveTo(Ctrl_Cpu *p_ctrl, HitTile *p_tile) {}


/*
 * Detach a tile
 */
void Ctrl_Cpu_Detach(Ctrl_Cpu *p_ctrl, HitTile *p_tile){
	
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_DETACH;

	Ctrl_Cpu_AddTaskToQueue(p_ctrl, task);

	sem_wait(&(p_ctrl->sem_sync));
}


/*
 * Move a tile from device (async version)
 */
void Ctrl_Cpu_MoveFrom(Ctrl_Cpu *p_ctrl, HitTile *p_tile) {}


/*
 * Create internally a data structure
 */
void Ctrl_Cpu_CreateInternal(Ctrl_Cpu *p_ctrl, HitTile *p_tile) {
	//TODO @arturo hay que convertir hitmap en estatica 
	// hit_tileAlloc(p_tile);
}


/*
 * Destroy Internal Function: recover is use dto choice between only free the memory or also take back the data 
 */
void Ctrl_Cpu_DestroyInternal(Ctrl_Cpu *p_ctrl, HitTile *p_tile) {
	//TODO @arturo hay que convertir hitmap en estatica 
	// hit_tileFree(*p_tile);
}


/*
 * WaitTile Function: block host thread execution until communications of tiles are finished
 */
void Ctrl_Cpu_WaitTile(Ctrl_Cpu *p_ctrl, HitTile *p_tile) {
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_WAITTILE;

	Ctrl_Cpu_AddTaskToQueue(p_ctrl, task);

	sem_wait(&(p_ctrl->sem_sync));
}


/**********************************
 * TASKS'S EVALUATION FUNCTIONS ***
 **********************************/


void Ctrl_Cpu_EvalTaskGlobalSync(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task) {
	sem_post(&(p_ctrl->sem_sync));
}

/* 
 * Evaluation of kernel launchs
 * Params:
 *      ctrl: Controller in charge of task.
 *      task: task that must be evaluated.
 */
void Ctrl_Cpu_EvalTaskKernelLaunch(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task) {
	Ctrl_Request request;
	request.cpu.n_cores = p_ctrl->n_cores; 

	p_task->pfn_kernel_wrapper(request, CTRL_TYPE_CPU, p_task->device_id, p_task->threads, p_task->p_arguments);
}


/* 
 * Evaluation of attach of tiles in controller
 * Params:
 *      cntrl: Controller in charge of task.
 *      run: task that must be evaluated.
 */
void Ctrl_Cpu_EvalTaskAttachTile(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task) {}


/**
 * Evaluation of detach of tiles from controller
 * Params:
 *      p_ctrl: Controller in charge of task.
 *      p_task: task that must be evaluated.
 */
void Ctrl_Cpu_EvalTaskDetachTile(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task) {
	sem_post(&(p_ctrl->sem_sync));
}


/* 
 * Evaluation of comunication of tile from host to device
 * Params:
 *      p_ctrl: Controller in charge of task.
 *      p_task: task that must be evaluated.
 */
void Ctrl_Cpu_EvalTaskMoveTo(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task) {}

/* 
 * Evaluation of comunication of tile from device to host
 * Params:
 *      cntrl: Controller in charge of task.
 *      run: task that must be evaluated.
 */
void Ctrl_Cpu_EvalTaskMoveFrom(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task) {}


/* 
 * Evaluation of synchronization with tile.
 * Params:
 *      cntrl: Controller in charge of task.
 *      run: task that must be evaluated.
 */
void Ctrl_Cpu_EvalTaskWaitTile(Ctrl_Cpu *p_ctrl, Ctrl_Task *p_task) {
	sem_post(&(p_ctrl->sem_sync));
}
