/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include "Architectures/Cuda/Ctrl_Cuda.h"

void Ctrl_Cuda_EvalTaskKernelLaunch(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskHostTaskLaunch(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskGlobalSync(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskCreateTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskAllocTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskFreeTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskMoveTo(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskMoveToInner(Ctrl_Cuda *p_ctrl, HitTile *p_tile);
void Ctrl_Cuda_EvalTaskMoveFrom(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_EvalTaskMoveFromInner(Ctrl_Cuda *p_ctrl, HitTile *p_tile);
void Ctrl_Cuda_EvalTaskWaitTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);

void Ctrl_Cuda_Sync(Ctrl_Cuda *p_ctrl);


void Ctrl_Cuda_NewThreadSetup(Ctrl_Cuda *p_ctrl) {
    CUDA_OP( cudaSetDevice(p_ctrl->device) );
}


void Ctrl_Cuda_Create(Ctrl_Cuda *p_ctrl, Ctrl_Policy policy, int device) {
	p_ctrl->policy = policy;
    p_ctrl->device = device;

	p_ctrl->p_tile_list_head = NULL;
	p_ctrl->p_tile_list_tail = NULL;

    CUDA_OP( cudaStreamCreateWithFlags(&(p_ctrl->stream_kernel), cudaStreamNonBlocking) );
    CUDA_OP( cudaStreamCreateWithFlags(&(p_ctrl->stream_host), cudaStreamNonBlocking) );

	CUDA_OP( cudaEventCreateWithFlags(&(p_ctrl->event_seq), cudaEventDisableTiming) );
}


void Ctrl_Cuda_Destroy(Ctrl_Cuda *p_ctrl) {
	p_ctrl->p_tile_list_head = NULL;
	p_ctrl->p_tile_list_tail = NULL;

	CUDA_OP( cudaStreamDestroy(p_ctrl->stream_kernel) );
    CUDA_OP( cudaStreamDestroy(p_ctrl->stream_host) );

	CUDA_OP( cudaEventDestroy(p_ctrl->event_seq) );
}


void Ctrl_Cuda_Sync(Ctrl_Cuda *p_ctrl) {
	omp_set_lock(p_ctrl->p_lock_first_ctrl);
	omp_unset_lock(p_ctrl->p_lock_ctrl);
	omp_set_lock(p_ctrl->p_lock_host);
	omp_unset_lock(p_ctrl->p_lock_first_ctrl);

	omp_set_lock(p_ctrl->p_lock_first_host);
	omp_unset_lock(p_ctrl->p_lock_host);
	omp_set_lock(p_ctrl->p_lock_ctrl);
	omp_unset_lock(p_ctrl->p_lock_first_host);
}

void Ctrl_Cuda_EvalTask(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	switch ( p_task->task_type ) {
		case CTRL_TASK_TYPE_KERNEL: 
			Ctrl_Cuda_EvalTaskKernelLaunch(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_HOST: 
			Ctrl_Cuda_EvalTaskHostTaskLaunch(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_GLOBALSYNC:
			Ctrl_Cuda_EvalTaskGlobalSync(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_ALLOCTILE:
			Ctrl_Cuda_EvalTaskAllocTile(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_FREETILE:
			Ctrl_Cuda_EvalTaskFreeTile(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_MOVETO:
			Ctrl_Cuda_EvalTaskMoveTo(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_MOVEFROM:
			Ctrl_Cuda_EvalTaskMoveFrom(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_WAITTILE:
			Ctrl_Cuda_EvalTaskWaitTile(p_ctrl, p_task);
			break;
	}
}


void Ctrl_Cuda_EvalTaskGlobalSync(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	CUDA_OP( cudaDeviceSynchronize() );
	Ctrl_Cuda_Sync(p_ctrl);
}


void Ctrl_Cuda_EvalTaskKernelLaunch(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	for (int i = 0; i < p_task->n_arguments; i++) {
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = (HitTile *)(p_task->pp_pointers[i]);
			KHitTile *p_ktile = (KHitTile *)((uint8_t *)p_task->p_arguments + p_task->p_displacements[i]);
			Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

			if (p_task->p_roles[i] != KERNEL_OUT) {
				if (p_tile_data->last_update == CTRL_TILE_LAST_UPDATE_HOST) {
					Ctrl_Cuda_EvalTaskMoveToInner(p_ctrl, p_tile);
				}
			}

			p_ktile->data = p_tile_data->p_device_data;

			CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_kernel, p_tile_data->kernel_last_write_event, 0) );
			CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_kernel, p_tile_data->offloading_last_write_event, 0) );

			if (p_task->p_roles[i] != KERNEL_IN) {
				CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_kernel, p_tile_data->kernel_last_read_event, 0) );
				CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_kernel, p_tile_data->offloading_last_read_event, 0) );
			}
		}
	}

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_kernel, p_ctrl->event_seq, 0) );
	}

	Ctrl_Request request;

	request.cuda.stream = &(p_ctrl->stream_kernel);
	
	p_task->pfn_kernel_wrapper(request, CUDA, p_task->device_id, p_task->threads, p_task->p_arguments);

	for (int i = 0; i < p_task->n_arguments; i++) {
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = (HitTile *)(p_task->pp_pointers[i]);
			Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

			if (p_task->p_roles[i] != KERNEL_IN) {
				p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_DEV;
				CUDA_OP( cudaEventRecord(p_tile_data->kernel_last_write_event, p_ctrl->stream_kernel) );
			}

			if (p_task->p_roles[i] != KERNEL_OUT) {
				CUDA_OP( cudaEventRecord(p_tile_data->kernel_last_read_event, p_ctrl->stream_kernel) );
			}
		}
	}

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaEventRecord(p_ctrl->event_seq, p_ctrl->stream_kernel) );
	}
}



void Ctrl_Cuda_EvalTaskHostTaskLaunch(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	for (int i = 0; i < p_task->n_arguments; i++) {
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = (HitTile *)(p_task->pp_pointers[i]);
			Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

			if (p_task->p_roles[i] != KERNEL_OUT) {
				if (p_tile_data->last_update == CTRL_TILE_LAST_UPDATE_DEV) {
					Ctrl_Cuda_EvalTaskMoveFromInner(p_ctrl, p_tile);
				}
			}

			CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_host, p_tile_data->offloading_last_read_event, 0) );
			CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_host, p_tile_data->host_last_write_event, 0) );

			if (p_task->p_roles[i] != KERNEL_IN) {
				p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST;

				CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_host, p_tile_data->offloading_last_write_event, 0) );
				CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_host, p_tile_data->host_last_read_event, 0) );
			}

		}
	}

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaStreamWaitEvent(p_ctrl->stream_host, p_ctrl->event_seq, 0) );
		CUDA_OP( cudaEventSynchronize(p_ctrl->event_seq) );
		p_task->pfn_hostTask_wrapper(p_task->p_arguments);
	} else {
		CUDA_OP( cudaLaunchHostFunc(p_ctrl->stream_host, p_task->pfn_hostTask_wrapper, p_task->p_arguments) );
	}


	for (int i = 0; i < p_task->n_arguments; i++) {
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = (HitTile *)(p_task->pp_pointers[i]);
			Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

			if (p_task->p_roles[i] != KERNEL_IN) {
				CUDA_OP( cudaEventRecord(p_tile_data->host_last_write_event, p_ctrl->stream_host) );
			}

			if (p_task->p_roles[i] != KERNEL_OUT) {
				CUDA_OP( cudaEventRecord(p_tile_data->host_last_read_event, p_ctrl->stream_host) );
			}
		}
	}

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaEventRecord(p_ctrl->event_seq, p_ctrl->stream_host) );
		CUDA_OP( cudaEventSynchronize(p_ctrl->event_seq) );
		Ctrl_Cuda_Sync(p_ctrl);
	}
}


void Ctrl_Cuda_EvalTaskCreateTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)malloc(sizeof(Ctrl_Cuda_Tile));

	p_tile_data->p_ctrl = p_ctrl;

	CUDA_OP( cudaEventCreateWithFlags(&(p_tile_data->kernel_last_read_event), cudaEventDisableTiming) );
	CUDA_OP( cudaEventCreateWithFlags(&(p_tile_data->kernel_last_write_event), cudaEventDisableTiming) );

	CUDA_OP( cudaEventCreateWithFlags(&(p_tile_data->offloading_last_read_event), cudaEventDisableTiming) );
	CUDA_OP( cudaEventCreateWithFlags(&(p_tile_data->offloading_last_write_event), cudaEventDisableTiming) );

	CUDA_OP( cudaEventCreateWithFlags(&(p_tile_data->host_last_read_event), cudaEventDisableTiming) );
	CUDA_OP( cudaEventCreateWithFlags(&(p_tile_data->host_last_write_event), cudaEventDisableTiming) );

	p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST_DEV;

	p_tile->ext = (void *)p_tile_data;
}


void Ctrl_Cuda_EvalTaskAllocTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	Ctrl_Cuda_EvalTaskCreateTile(p_ctrl, p_task);

	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

	Ctrl_Cuda_Tile_List *p_list_node = (Ctrl_Cuda_Tile_List *)malloc(sizeof(Ctrl_Cuda_Tile_List));
	
	p_list_node->p_prev = NULL;
	p_list_node->p_next = NULL;
	p_list_node->p_tile_ext = p_tile_data;
	p_tile_data->p_tile_elem = p_list_node;

	if (p_ctrl->p_tile_list_tail != NULL){
		p_ctrl->p_tile_list_tail->p_next = p_list_node;
		p_list_node->p_prev = p_ctrl->p_tile_list_tail;
		p_ctrl->p_tile_list_tail = p_list_node;
	} else {
		p_ctrl->p_tile_list_head = p_ctrl->p_tile_list_tail = p_list_node;
	}

	CUDA_OP( cudaStreamCreateWithFlags(&(p_tile_data->stream), cudaStreamNonBlocking) );

	CUDA_OP( cudaMalloc(&(p_tile_data->p_device_data),((size_t)(p_tile->acumCard)) * (p_tile->baseExtent)) );
	CUDA_OP( cudaMallocHost(&(p_tile->data), ((size_t)(p_tile->acumCard)) * (p_tile->baseExtent)) );
	
	p_tile->memPtr = p_tile->data;

	if ( (p_tile->memStatus == HIT_MS_NOT_OWNER) || (p_tile->memStatus == HIT_MS_NOMEM) ) {
		for (int i = 0; i < hit_shapeDims(p_tile->shape); i++) {
			p_tile->qstride[i] = 1;
		}
		
		int cardinality = 1;
		p_tile->origAcumCard[hit_shapeDims(p_tile->shape)] = 1;

		for (int i = (hit_shapeDims(p_tile->shape) - 1); i >= 0; i--) {
			cardinality = cardinality * p_tile->card[i];
			p_tile->origAcumCard[i] = cardinality;
		}
		p_tile->acumCard = p_tile->origAcumCard[0];
	}

	p_tile->memStatus = HIT_MS_OWNER;
	
	Ctrl_Cuda_Sync(p_ctrl);
}



void Ctrl_Cuda_EvalTaskFreeTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

	CUDA_OP( cudaFreeHost(p_tile->data) );

	if (p_tile->memStatus == HIT_MS_OWNER) {
		p_tile->memPtr = NULL;
		p_tile->data = NULL;
		p_tile->memStatus = HIT_MS_NOMEM;
	}

	CUDA_OP( cudaFree(p_tile_data->p_device_data) );

	CUDA_OP( cudaEventDestroy(p_tile_data->kernel_last_read_event) );
	CUDA_OP( cudaEventDestroy(p_tile_data->kernel_last_write_event) );

	CUDA_OP( cudaEventDestroy(p_tile_data->offloading_last_read_event) );
	CUDA_OP( cudaEventDestroy(p_tile_data->offloading_last_write_event) );

	CUDA_OP( cudaEventDestroy(p_tile_data->host_last_read_event) );
	CUDA_OP( cudaEventDestroy(p_tile_data->host_last_write_event) );

    CUDA_OP( cudaStreamDestroy(p_tile_data->stream) );

	if (p_tile_data->p_tile_elem->p_prev != NULL) {
		if (p_tile_data->p_tile_elem->p_next == NULL) {
			p_tile_data->p_tile_elem->p_prev->p_next = NULL;
			p_ctrl->p_tile_list_tail = p_tile_data->p_tile_elem->p_prev;
		} else {
			p_tile_data->p_tile_elem->p_prev->p_next = p_tile_data->p_tile_elem->p_next;
		}
	}

	if (p_tile_data->p_tile_elem->p_next != NULL) {
		if (p_tile_data->p_tile_elem->p_prev == NULL) {
			p_tile_data->p_tile_elem->p_next->p_prev = NULL;
			p_ctrl->p_tile_list_head = p_tile_data->p_tile_elem->p_next;
		} else {
			p_tile_data->p_tile_elem->p_next->p_prev = p_tile_data->p_tile_elem->p_prev;
		}
	}

	p_tile->ext = NULL;
	p_tile_data->p_ctrl = NULL;
	
	p_tile_data->p_tile_elem->p_tile_ext = NULL;
	p_tile_data->p_tile_elem->p_next = NULL;
	p_tile_data->p_tile_elem->p_prev = NULL;

	free(p_tile_data->p_tile_elem);
	free(p_tile_data);

	Ctrl_Cuda_Sync(p_ctrl);
}


void Ctrl_Cuda_EvalTaskMoveTo(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

	if (p_tile_data->last_update != CTRL_TILE_LAST_UPDATE_HOST_DEV) {
		Ctrl_Cuda_EvalTaskMoveToInner(p_ctrl, p_tile);
	}
	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaEventSynchronize(p_ctrl->event_seq) );
		Ctrl_Cuda_Sync(p_ctrl);
	}
}


void Ctrl_Cuda_EvalTaskMoveToInner(Ctrl_Cuda *p_ctrl, HitTile *p_tile) {
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->kernel_last_read_event, 0) );
	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->kernel_last_write_event, 0) );
	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->offloading_last_read_event, 0) );
	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->host_last_write_event, 0) );

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_ctrl->event_seq, 0) );
	}

	CUDA_OP( 
		cudaMemcpyAsync(
			p_tile_data->p_device_data, 
			p_tile->data, 
			((size_t)(p_tile->acumCard)) * (p_tile->baseExtent), 
			cudaMemcpyHostToDevice, 
			p_tile_data->stream
		)	
	);

	CUDA_OP( cudaEventRecord(p_tile_data->offloading_last_write_event, p_tile_data->stream) );
	
	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaEventRecord(p_ctrl->event_seq, p_tile_data->stream) );
	}

	p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST_DEV;
}


void Ctrl_Cuda_EvalTaskMoveFrom(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

	if (p_tile_data->last_update != CTRL_TILE_LAST_UPDATE_HOST_DEV) {
		Ctrl_Cuda_EvalTaskMoveFromInner(p_ctrl, p_tile);
	}

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaEventSynchronize(p_ctrl->event_seq) );
		Ctrl_Cuda_Sync(p_ctrl);
	}
}


void Ctrl_Cuda_EvalTaskMoveFromInner(Ctrl_Cuda *p_ctrl, HitTile *p_tile) {
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->kernel_last_write_event, 0) );
	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->offloading_last_write_event, 0) );
	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->host_last_read_event, 0) );
	CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_tile_data->host_last_write_event, 0) );

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaStreamWaitEvent(p_tile_data->stream, p_ctrl->event_seq, 0) );
	}

	CUDA_OP( 
		cudaMemcpyAsync(
			p_tile->data, 
			p_tile_data->p_device_data, 
			((size_t)(p_tile->acumCard)) * (p_tile->baseExtent), 
			cudaMemcpyDeviceToHost, 
			p_tile_data->stream
		)	
	);

	CUDA_OP( cudaEventRecord(p_tile_data->offloading_last_read_event, p_tile_data->stream) );

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		CUDA_OP( cudaEventRecord(p_ctrl->event_seq, p_tile_data->stream) );
	}

	p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST_DEV;
}

void Ctrl_Cuda_EvalTaskWaitTile(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_Cuda_Tile *p_tile_data = (Ctrl_Cuda_Tile *)(p_tile->ext);

	CUDA_OP( cudaEventSynchronize(p_tile_data->kernel_last_read_event) );
	CUDA_OP( cudaEventSynchronize(p_tile_data->kernel_last_write_event) );

	CUDA_OP( cudaEventSynchronize(p_tile_data->offloading_last_read_event) );
	CUDA_OP( cudaEventSynchronize(p_tile_data->offloading_last_write_event) );

	CUDA_OP( cudaEventSynchronize(p_tile_data->host_last_read_event) );
	CUDA_OP( cudaEventSynchronize(p_tile_data->host_last_write_event) );

	Ctrl_Cuda_Sync(p_ctrl);
}
