/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include <stdbool.h>
#include <stdio.h>

#include "Core/Ctrl_KHitTile.h"
#include "Core/Ctrl_Request.h"
#include "Core/Ctrl_Core.h"

#include "Kernel/Ctrl_KernelArgs.h"

#include "Architectures/XPhi/Ctrl_XPhi.h"
#include "Architectures/XPhi/Ctrl_XPhi_Request.h"
#include "Architectures/XPhi/Ctrl_XPhi_Tile.h"



/*************************************************************
 ******** Prototypes of private functions ********************
 *************************************************************/

void Ctrl_XPhi_ConsumeTaskQueue(Ctrl_XPhi *p_ctrl);

/*************************************************************
 ******** Prototypes of tasks' evaluation functions **********
 *************************************************************/
void Ctrl_XPhi_EvalTaskKernelLaunch(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task);
void Ctrl_XPhi_EvalTaskGlobalSync(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task);
void Ctrl_XPhi_EvalTaskAttachTile(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task, int isInternal);
void Ctrl_XPhi_EvalTaskAllocPinned(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task);
void Ctrl_XPhi_EvalTaskFreePinned(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task);
void Ctrl_XPhi_EvalTaskDetachTile(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task);
void Ctrl_XPhi_EvalTaskCreateInternal(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task);
void Ctrl_XPhi_EvalTaskMoveTo(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task, bool from_attach);
void Ctrl_XPhi_EvalTaskMoveFrom(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task, bool from_dettach);
void Ctrl_XPhi_EvalTaskWaitTile(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task);

/********************************************
 ******** GPU Controller functions **********
 ********************************************/

/*
 * Create the controller and its corresponding variables
 * Params:
 *      cntrl: Controller to be created.
 *      device: GPU used by this controller
 */
void Ctrl_XPhi_Create(Ctrl_XPhi *p_ctrl, int n_xphi, Ctrl_Policy policy) {
	
	if(policy != CTRL_POLICY_DEFAULT) {
		p_ctrl->policy = policy;
	} else {
		if ( getenv("CTRL_ASYNC_MODE") ) {
			if (atoi( getenv("CTRL_ASYNC_MODE") ) > 0) {
				p_ctrl->policy = CTRL_POLICY_ASYNC;
			} else {
				p_ctrl->policy = CTRL_POLICY_SYNC;
			}
		} else {
			p_ctrl->policy = CTRL_POLICY_SYNC;
		}
	}

	p_ctrl->n_task = 0;

	/* Allocate the controller */
	Ctrl_TaskQueue_Init(&(p_ctrl->task_queue));

	p_ctrl->n_xphi = n_xphi;
	int def_signal = 0;
	p_ctrl->default_signal = def_signal;
	#pragma offload_transfer target(mic:n_xphi) signal(def_signal)
	p_ctrl->n_signal = 11;

	// p_ctrl->n_engines = (uint32_t) _Offload_number_of_devices();
	p_ctrl->stream = _Offload_stream_create(p_ctrl->n_xphi, 255);

	/* Initialize the data structure */

	// omp_init_lock(&(p_ctrl->destroy_ctrl));
	// omp_init_lock(&(p_ctrl->global_sync));
	omp_init_lock(&(p_ctrl->lock_queue));
	
	sem_init(&(p_ctrl->sem_destroy), 0, 1);
	sem_init(&(p_ctrl->sem_sync), 0, 0);

	sem_wait(&(p_ctrl->sem_destroy));

	//TODO @arturo afinidades
	#pragma omp task
	{
		{
			{
				Ctrl_XPhi_ConsumeTaskQueue(p_ctrl);

				_Offload_stream_destroy(p_ctrl->n_xphi, p_ctrl->stream);	
				
				omp_set_lock(&(p_ctrl->lock_queue));
				Ctrl_TaskQueue_Clean(&(p_ctrl->task_queue));
				omp_unset_lock(&(p_ctrl->lock_queue));

				omp_destroy_lock(&(p_ctrl->lock_queue));

				sem_post(&(p_ctrl->sem_destroy));
			}
		}
	}
}


/*
 * Synchornous policy to consume queue of tasks 
 * Params:
 *      ctrl: instance of Controller.
 */
void Ctrl_XPhi_ConsumeTaskQueue(Ctrl_XPhi *p_ctrl) {
	
	int task_done = 0;
	bool finish = false;

	while (!finish) {

		omp_set_lock(&(p_ctrl->lock_queue));
		if ( !(Ctrl_TaskQueue_IsEmpty(&(p_ctrl->task_queue))) ) {
			
			Ctrl_Task *p_task = Ctrl_TaskQueue_Pop( &(p_ctrl->task_queue) );
			omp_unset_lock(&(p_ctrl->lock_queue));

#ifdef CTRL_DEBUG
		printf("\n");
		printf("Ctrl_XPhi_ConsumeTaskQueue: %i ", p_task->task_type);
			switch ( p_task->task_type ) {
				case CTRL_TASK_TYPE_KERNEL: 
					printf("CTRL_TASK_TYPE_KERNEL");
					break;
				case CTRL_TASK_TYPE_DESTROYCNTRL:
					printf("CTRL_TASK_TYPE_DESTROYCNTRL");
					break;
				case CTRL_TASK_TYPE_GLOBALSYNC:
					printf("CTRL_TASK_TYPE_GLOBALSYNC");
					break;
                                case CTRL_TASK_TYPE_ALLOCPINNED:
                                        printf("CTRL_TASK_TYPE_ALLOCPINNED");
                                        break;
                                case CTRL_TASK_TYPE_FREEPINNED:
					printf("CTRL_TASK_TYPE_FREEPINNED");
                                        break;
				case CTRL_TASK_TYPE_ATTACH:
					printf("CTRL_TASK_TYPE_ATTACH");
					break;
				case CTRL_TASK_TYPE_DETACH: 
					printf("CTRL_TASK_TYPE_DETACH");
					break;
				case CTRL_TASK_TYPE_MOVETO:
					printf("CTRL_TASK_TYPE_MOVETO");
					break;
				case CTRL_TASK_TYPE_MOVEFROM:
					printf("CTRL_TASK_TYPE_MOVEFROM");
					break;
				case CTRL_TASK_TYPE_WAITTILE:
					printf("CTRL_TASK_TYPE_WAITTILE");
					break;
				case CTRL_TASK_TYPE_CREATEINTERNAL:
					printf("CTRL_TASK_TYPE_CREATEINTERNAL");
					break;
				case CTRL_TASK_TYPE_DESTROYINTERNAL:
					printf("CTRL_TASK_TYPE_DESTROYINTERNAL");
					break;
			}
			printf("\n");
#endif
		
			// Evaluation of task
			switch ( p_task->task_type ) {
				case CTRL_TASK_TYPE_KERNEL: 
					Ctrl_XPhi_EvalTaskKernelLaunch(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_DESTROYCNTRL:
					finish = true;
					break;
				case CTRL_TASK_TYPE_GLOBALSYNC:
					Ctrl_XPhi_EvalTaskGlobalSync(p_ctrl, p_task);
					break;
                                case CTRL_TASK_TYPE_ALLOCPINNED:
                                        Ctrl_XPhi_EvalTaskAllocPinned(p_ctrl, p_task);
                                        break;
                                case CTRL_TASK_TYPE_FREEPINNED:
                                        Ctrl_XPhi_EvalTaskFreePinned(p_ctrl, p_task);
                                        break;
				case CTRL_TASK_TYPE_ATTACH:
					Ctrl_XPhi_EvalTaskAttachTile(p_ctrl, p_task, false);
					break;
				case CTRL_TASK_TYPE_DETACH: 
					Ctrl_XPhi_EvalTaskDetachTile(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_MOVETO:
					Ctrl_XPhi_EvalTaskMoveTo(p_ctrl, p_task, false);
					break;
				case CTRL_TASK_TYPE_MOVEFROM:
					Ctrl_XPhi_EvalTaskMoveFrom(p_ctrl, p_task, false);
					break;
				case CTRL_TASK_TYPE_WAITTILE:
					Ctrl_XPhi_EvalTaskWaitTile(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_CREATEINTERNAL:
					//Ctrl_XPhi_EvalTaskAttachTile(p_ctrl, p_task, true);
					Ctrl_XPhi_EvalTaskCreateInternal(p_ctrl, p_task);
					break;
				case CTRL_TASK_TYPE_DESTROYINTERNAL:
					Ctrl_XPhi_EvalTaskDetachTile(p_ctrl, p_task);
					break;
			}
			task_done++;
			free(p_task);
		} else {
			omp_unset_lock(&(p_ctrl->lock_queue));
		}
	}
}


/*********************************
 * CONTROLLER EXTERNAL FUNCTIONS *
 *********************************/

//Add task to the cntrl
void Ctrl_XPhi_AddTaskToQueue(Ctrl_XPhi *p_ctrl, Ctrl_Task task) {
	task.task_id = p_ctrl->n_task;

	omp_set_lock(&(p_ctrl->lock_queue));
	Ctrl_TaskQueue_Push(&(p_ctrl->task_queue), task);
	omp_unset_lock(&(p_ctrl->lock_queue));

	p_ctrl->n_task += 1;
}

void Ctrl_XPhi_LaunchKernel(Ctrl_XPhi *p_ctrl, Ctrl_Task task) {
	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);
}


/*
 * Destroy a Controller
 */
void Ctrl_XPhi_Destroy(Ctrl_XPhi *p_ctrl) {
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_DESTROYCNTRL;
	
	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	sem_wait(&(p_ctrl->sem_destroy));
	sem_destroy(&(p_ctrl->sem_destroy));
	sem_destroy(&(p_ctrl->sem_sync));
	//FIXME: Falta algo aquí?
}


void Ctrl_XPhi_GlobalSync(Ctrl_XPhi *p_ctrl){
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_GLOBALSYNC;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);
	
	sem_wait(&(p_ctrl->sem_sync));
}


/*
 * Attach a tile
 */
void Ctrl_XPhi_Attach(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {
	
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_ATTACH;

	task.p_arguments = (HitTile *)p_tile;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		sem_wait(&(p_ctrl->sem_sync));
	}
}

void Ctrl_XPhi_AllocPinned(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {
        Ctrl_Task task = CTRL_TASK_NULL;
        task.task_type = CTRL_TASK_TYPE_ALLOCPINNED;

        task.p_arguments = (HitTile *)p_tile;

        Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

        sem_wait(&(p_ctrl->sem_sync));
}

void Ctrl_XPhi_FreePinned(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {
        Ctrl_Task task = CTRL_TASK_NULL;
        task.task_type = CTRL_TASK_TYPE_FREEPINNED;

        task.p_arguments = (HitTile *)p_tile;

        Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

        sem_wait(&(p_ctrl->sem_sync));
}

/*
 * Move a tile to device
 */
void Ctrl_XPhi_MoveTo(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {

	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_MOVETO;

	task.p_arguments = (HitTile *)p_tile;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		sem_wait(&(p_ctrl->sem_sync));
	}
}


/*
 * Detach a tile
 */
void Ctrl_XPhi_Detach(Ctrl_XPhi *p_ctrl, HitTile *p_tile){
	
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_DETACH;

	task.p_arguments = (HitTile *)p_tile;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	sem_wait(&(p_ctrl->sem_sync));
}


/*
 * Move a tile from device (async version)
 */
void Ctrl_XPhi_MoveFrom(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_MOVEFROM;

	task.p_arguments = (HitTile *)p_tile;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		sem_wait(&(p_ctrl->sem_sync));
	}
}


/*
 * Create internally a data structure
 */
void Ctrl_XPhi_CreateInternal(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_CREATEINTERNAL;

	task.p_arguments = (HitTile *)p_tile;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		sem_wait(&(p_ctrl->sem_sync));
	}
}


/*
 * Destroy Internal Function: recover is use dto choice between only free the memory or also take back the data 
 */
void Ctrl_XPhi_DestroyInternal(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_DESTROYINTERNAL;

	task.p_arguments = (HitTile *)p_tile;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		sem_wait(&(p_ctrl->sem_sync));
	}
}


/*
 * WaitTile Function: block host thread execution until communications of tiles are finished
 */
void Ctrl_XPhi_WaitTile(Ctrl_XPhi *p_ctrl, HitTile *p_tile) {
	Ctrl_Task task = CTRL_TASK_NULL;
	task.task_type = CTRL_TASK_TYPE_WAITTILE;

	task.p_arguments = (HitTile *)p_tile;

	Ctrl_XPhi_AddTaskToQueue(p_ctrl, task);

	sem_wait(&(p_ctrl->sem_sync));
}


/**********************************
 * TASKS'S EVALUATION FUNCTIONS ***
 **********************************/


void Ctrl_XPhi_EvalTaskGlobalSync(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task) {
	//TODO @eduardo esperar por el ultimo signal?
	int n_xphi = p_ctrl->n_xphi;
	int last_signal = p_ctrl->n_signal;

	//#pragma offload_wait target(mic:n_xphi) wait(last_signal)
	#pragma offload_wait target(mic:n_xphi) stream(0)

	sem_post(&(p_ctrl->sem_sync));
}

void Ctrl_XPhi_EvalTaskAllocPinned(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task)
{
        //Ctrl_OpenCLGpu_EvalTaskCreateTile(p_ctrl, p_task, true);
        HitTile *p_tile = (HitTile *)(p_task->p_arguments);
	//TODO@eduardo: En este punto, no se ha hecho aún un malloc de Ctrl_XPhi_Tile, con lo que no existe. Mirar el modelo de waxa, arreglar y descomentar las dos lineas de abajo:
        //Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)(p_tile->ext);

        //p_tile_data->isPinned = true;

	// @arturo Ago 2015: New allocP interface
	p_tile->data = _mm_malloc(p_tile->acumCard * p_tile->baseExtent, 64);
#ifdef CTRL_DEBUG
	printf("malloc %p\n", p_tile->data);
#endif
        p_tile->memPtr = p_tile->data;

	/* 3. UPDATES FOR SHADOW COPIES, OR ALLOCATION OF SELECTIONS OF NO-MEMORY VARIABLES */
        if ( (p_tile->memStatus == HIT_MS_NOT_OWNER) || (p_tile->memStatus == HIT_MS_NOMEM) ) {
		/* 3.1. NEW STRIDES TO ACCESS ARE ALWAYS 1 */
                for (int i = 0; i < hit_shapeDims(p_tile->shape); i++) {
                        p_tile->qstride[i] = 1;
                }

		/* 3.2. UPDATE ORIGINAL ACUMULATED CARDINALITIES, NOW IT HAS ITS OWN MEMORY */
		//Code below extracted from the body of hit_tileUpdateAcumCards(p_tile);
                p_tile->origAcumCard[hit_shapeDims(p_tile->shape)] = 1;
                int cardinality = 1;

                for (int i = (hit_shapeDims(p_tile->shape) - 1); i >= 0; i--) {
                        cardinality = cardinality * p_tile->card[i];
                        p_tile->origAcumCard[i] = cardinality;
                }
                p_tile->acumCard = p_tile->origAcumCard[0];
        }

        p_tile->memStatus = HIT_MS_OWNER;

        sem_post(&(p_ctrl->sem_sync));
}

void Ctrl_XPhi_EvalTaskFreePinned(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task)
{

        HitTile *p_tile = (HitTile *)(p_task->p_arguments);
	//TODO @eduardo: At this point, p_tile->ext has been destroyed and assigned NULL by
	// the Detach function. Talk about this with waxa and maybe uncomment the following
	// lines. 
	//Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)(p_tile->ext);

	//Wait for all tasks enqueued on the stream to complete before freeing the variable.
	//int n_xphi = p_ctrl->n_xphi;
	//_Offload_stream  stream = p_tile_data->stream; 
	//#pragma offload_wait target(mic:n_xphi) stream(stream)

	_mm_free(p_tile->data);

        p_tile->ext = NULL;
        //p_tile_data->p_ctrl = NULL;

        //free(p_tile_data);

        if (p_tile->memStatus == HIT_MS_OWNER) {
                p_tile->memPtr = NULL;
                p_tile->data = NULL;
                p_tile->memStatus = HIT_MS_NOMEM;
        }
        sem_post(&(p_ctrl->sem_sync));
}

/* 
 * Evaluation of kernel launchs
 * Params:
 *      ctrl: Controller in charge of task.
 *      task: task that must be evaluated.
 */
void Ctrl_XPhi_EvalTaskKernelLaunch(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task) {
	//TODO @eduardo lanzar la funcion real

	//int wait_signal = p_ctrl->n_signal++;
	int new_signal = p_ctrl->n_signal++; //Signal que tiene que generar el kernel

	int n_xphi = p_ctrl->n_xphi;

	//Prescan valid signals
	int n_list_signals = 0;
	/*for (int i = 0; i < p_task->n_arguments; i++)
	{
		switch(p_task->p_roles[i])
		{
			case KERNEL_IN:   n_list_signals += 1; break;
			case KERNEL_OUT:  n_list_signals += 2; break;
			case KERNEL_IO:   n_list_signals += 2; break;
			default: break;
		}
	}
	int *p_list_signals = (int *)malloc(n_list_signals*sizeof(int));*/

	#define MAX_SIGNALS 20
	int p_list_signals[MAX_SIGNALS];

	for (int i = 0; i < p_task->n_arguments; i++) {

//		printf("n_list_signals %i\n", n_list_signals);
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = p_task->pp_pointers[i];
			Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile*)(p_tile->ext);
			KHitTile *p_ktile = ((uint8_t *)p_task->p_arguments + p_task->p_displacements[i]);

                        p_ktile->data = p_tile->data;
#ifdef CTRL_DEBUG
			printf("DATOS p_ktile card %i %i %i\n", p_ktile->card[0], p_ktile->card[1], p_ktile->card[2]);
#endif
	
			if (p_task->p_roles[i] == KERNEL_IN) {
#ifdef CTRL_DEBUG
				printf("IN p_ktile->data %p data[0] %f\n", p_ktile->data, ((float *)(p_ktile->data))[0]);
#endif
//				printf("Pacaaa %i\n", p_tile_data->signal_write);
				p_list_signals[n_list_signals] = p_tile_data->signal_write;
				n_list_signals++;
				(p_tile_data->signal_read) = new_signal;
//				printf("Pacaaa %i %i (total signals = %i)\n", p_list_signals[n_list_signals - 1], new_signal, n_list_signals - 1);
			}
			if (p_task->p_roles[i] == KERNEL_OUT) {
#ifdef CTRL_DEBUG
				printf("OUT p_ktile->data %p\n", p_ktile->data);
#endif
//				printf("Pacaaa OUT w %i r %i\n", p_tile_data->signal_write, p_tile_data->signal_read);
				p_list_signals[n_list_signals] = p_tile_data->signal_write;
				n_list_signals++;
				p_list_signals[n_list_signals] = p_tile_data->signal_read;
				n_list_signals++;
				(p_tile_data->signal_write) = new_signal;
//				printf("Pacaaa OUT w %i r %i write %i (total signals = %i)\n", p_list_signals[n_list_signals - 2], p_list_signals[n_list_signals - 1], new_signal, n_list_signals - 1);
			}
			if (p_task->p_roles[i] == KERNEL_IO) {
				p_list_signals[n_list_signals] = p_tile_data->signal_write;
				n_list_signals++;
				p_list_signals[n_list_signals] = p_tile_data->signal_read;
				n_list_signals++;
				(p_tile_data->signal_read) = new_signal;
				(p_tile_data->signal_write) = new_signal;
			}
		}
	}
//	printf("n_list_signals %i\n", n_list_signals);

	if (n_list_signals != 0) {
		for(int i = n_list_signals; i < MAX_SIGNALS; i++) {
			p_list_signals[i] = p_ctrl->default_signal;
		}
		/*#pragma offload_transfer target(mic:n_xphi) wait( \
			p_list_signals[0], p_list_signals[1], p_list_signals[2], p_list_signals[3], p_list_signals[4], \
			p_list_signals[5], p_list_signals[6], p_list_signals[7], p_list_signals[8], p_list_signals[9], \
			p_list_signals[10], p_list_signals[11], p_list_signals[12], p_list_signals[13], p_list_signals[14], \
			p_list_signals[15], p_list_signals[16], p_list_signals[17], p_list_signals[18], p_list_signals[19] \
			) signal(wait_signal)*/
	}


	//TODO @waxa quitar lista de signal del request cambiarla por el nuevo signal (wait_signal)

	Ctrl_Request request;
	request.xphi.stream_handler = p_ctrl->stream;
	request.xphi.new_signal = new_signal;
	request.xphi.n_signals = n_list_signals;
	request.xphi.p_signals = p_list_signals;
	request.xphi.default_signal = p_ctrl->default_signal;

#ifdef CTRL_DEBUG
	KHitTile *algo1 = (KHitTile*)(p_task->p_arguments);
	KHitTile algo = *(algo1);
	//printf("algo %p data = %p card = %i %i %i\n", algo, algo->data, algo->card[0], algo->card[1], algo->card[2]);
	printf("algo %p data = %p card = %i %i %i\n", &algo, algo.data, algo.card[0], algo.card[1], algo.card[2]);
	printf("primer dst = %p, k_dst = %p, t_dst = %p, orig_dst = %p\n", (KHitTile*)(p_task->p_arguments), *((KHitTile*) (p_task->p_arguments)), (*((KHitTile*) (p_task->p_arguments))).data, (*((KHitTile*) (p_task->p_arguments))).origAcumCard);
#endif
	p_task->pfn_kernel_wrapper(request, CTRL_TYPE_XPHI, p_task->device_id, p_task->threads, p_task->p_arguments);
}


/* 
 * Evaluation of attach of tiles in controller
 * Params:
 *      cntrl: Controller in charge of task.
 *      run: task that must be evaluated.
 */
void Ctrl_XPhi_EvalTaskAttachTile(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task, int isInternal) {
	HitTile *p_tile = (HitTile *)(p_task->p_arguments);

	Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)malloc(sizeof(Ctrl_XPhi_Tile));
	p_tile_data->p_ctrl = p_ctrl;

	p_tile_data->signal_read = p_ctrl->n_signal++;
	p_tile_data->signal_write = p_tile_data->signal_read;

	p_tile_data->stream = _Offload_stream_create(p_ctrl->n_xphi, 255);
	p_tile_data->isInternal = isInternal;

	//LEO always needs memory to be allocated on the host, wether or not the variable is internal
	if (isInternal)
		p_tile->data = malloc(p_tile->acumCard*p_tile->baseExtent);

	p_tile->ext = (void *)p_tile_data;

	int n_xphi = p_ctrl->n_xphi;
	char *p_data = (char*) p_tile->data;
	size_t data_size = p_tile->acumCard * p_tile->baseExtent;

	int init_signal = p_tile_data->signal_write;
	_Offload_stream  stream = p_tile_data->stream; 
	
#ifdef CTRL_DEBUG
	printf("Ctrl_XPhi_EvalTaskAttachTile offload_trans nocopy p_tile: %p p_data: %p length: %i\n", p_tile, p_data, data_size); fflush(stdout);
	printf("Ctrl_XPhi_EvalTaskAttachTile signal_read = %i, signal_write = %i, signal = %i", p_tile_data->signal_read, p_tile_data->signal_write, init_signal); fflush(stdout);
	printf("\n"); fflush(stdout);
#endif

#ifndef CTRL_DEBUG
	#pragma offload_transfer target(mic) nocopy(p_data: length(data_size) align(64) alloc_if(true) free_if(false)) stream(stream) signal(init_signal)
#else
	#pragma offload target(mic) nocopy(p_data: length(data_size) alloc_if(true) free_if(false)) stream(stream) signal(init_signal)
	{
	    printf("Ctrl_XPhi_EvalTaskAttachTile MIC p_data: %p\n", p_data);
	}
#endif
	
	if (p_ctrl->policy == CTRL_POLICY_SYNC){
		//WRONG RESULTS OCCASIONALLY
		//#pragma offload_wait target(mic) stream(stream) wait(init_signal)
		//SIGSEV
		//#pragma offload_wait target(mic:n_xphi) wait(init_signal)
		#pragma offload_wait target(mic) stream(0)
		}


	if(! isInternal)
		Ctrl_XPhi_EvalTaskMoveTo(p_ctrl, p_task, true);
	
	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		//printf("SYNC\n");
		//#pragma offload_wait target(mic) stream(stream) wait(init_signal)
		#pragma offload_wait target(mic:n_xphi)  wait(init_signal)
		//#pragma offload_wait target(mic) stream(0)
		sem_post(&(p_ctrl->sem_sync));
	}
}


void Ctrl_XPhi_EvalTaskCreateInternal(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_arguments);

	//TODO @eduardo esto se elimina cuando se pase a usar COI
	p_tile->data = _mm_malloc(p_tile->acumCard * p_tile->baseExtent, 64);
	//p_tile->data = malloc(p_tile->acumCard * p_tile->baseExtent);

	Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)malloc(sizeof(Ctrl_XPhi_Tile));
	p_tile_data->p_ctrl = p_ctrl;

	//V1
	//p_tile_data->signal_read = p_ctrl->default_signal;//p_ctrl->n_signal++;
	//p_tile_data->signal_write = p_ctrl->default_signal;
	//V2
	//p_tile_data->signal_read = p_ctrl->default_signal;
	//p_tile_data->signal_write = p_ctrl->n_signal++;
	//V3: You can't neither read nor write until the memory is allocated on the MIC (this operation)
	p_tile_data->signal_read = p_ctrl->n_signal++;
	p_tile_data->signal_write = p_tile_data->signal_read;

	p_tile_data->isInternal = true;

	p_tile->ext = (void *)p_tile_data;

	int n_xphi = p_ctrl->n_xphi;
	char* p_data = (char*) p_tile->data;
	size_t data_size = p_tile->acumCard * p_tile->baseExtent;

	int init_signal = p_tile_data->signal_write;
	
#ifdef CTRL_DEBUG
	printf("Ctrl_XPhi_EvalTaskCreateInternal offload_trans nocopy p_tile: %p p_data: %p length: %i\n", p_tile, p_data, data_size); fflush(stdout);
	printf("Ctrl_XPhi_EvalTaskCreateInternal signal_read = %i, signal_write = %i, signal = %i", p_tile_data->signal_read, p_tile_data->signal_write, init_signal); fflush(stdout);
	printf("\n"); fflush(stdout);
#endif
	#pragma offload_transfer target(mic:n_xphi) nocopy(p_data: length(data_size) align(64) alloc_if(true) free_if(false)) signal(init_signal)

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		#pragma offload_wait target(mic:n_xphi)  wait(init_signal)
		sem_post(&(p_ctrl->sem_sync));
	}
}

/**
 * Evaluation of detach of tiles from controller
 * Params:
 *      p_ctrl: Controller in charge of task.
 *      p_task: task that must be evaluated.
 */
void Ctrl_XPhi_EvalTaskDetachTile(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_arguments);
	Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)(p_tile->ext);

	if (!p_tile_data->isInternal) {
		Ctrl_XPhi_EvalTaskMoveFrom(p_ctrl, p_task, true);
	}

	int n_xphi = p_ctrl->n_xphi;
	int read_signal = p_tile_data->signal_read;
	int new_signal = p_ctrl->n_signal++;

	char* data = (char*) p_tile->data;
	size_t data_size = p_tile->acumCard * p_tile->baseExtent;
/*	printf("offfload_transfer wait %i\n", read_signal);fflush(stdout);
	printf("offfload_transfer wait %i\n", read_signal);fflush(stdout);
	printf("offfload_transfer wait %i\n", read_signal);fflush(stdout);
	printf("offfload_transfer wait %i\n", read_signal);fflush(stdout);*/
	#pragma offload_transfer target(mic:n_xphi) nocopy(data: length(data_size) align(64) alloc_if(false) free_if(true)) wait(read_signal) signal(new_signal)
	//printf("AAAAAAAAAofffload_transfer wait %i\n", read_signal);fflush(stdout);

	if (p_tile_data->isInternal) {
		_mm_free(p_tile->data);
		//free(p_tile->data);
	} else {
		_Offload_stream_destroy(p_ctrl->n_xphi, p_tile_data->stream);
	}

	p_tile_data->p_ctrl = NULL;
	free(p_tile_data);
	p_tile->ext = NULL;

	#pragma offload_wait target(mic:n_xphi)  wait(new_signal)

	sem_post(&(p_ctrl->sem_sync));
}


/* 
 * Evaluation of comunication of tile from host to device
 * Params:
 *      p_ctrl: Controller in charge of task.
 *      p_task: task that must be evaluated.
 */
void Ctrl_XPhi_EvalTaskMoveTo(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task, bool from_attach) {

	HitTile *p_tile = (HitTile *)(p_task->p_arguments);
	Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)(p_tile->ext);

	int n_xphi = p_ctrl->n_xphi;
	char* p_data = (char*) p_tile->data;
	size_t data_size = p_tile->acumCard * p_tile->baseExtent;

	_Offload_stream stream = p_tile_data->stream;

	int read_signal = p_tile_data->signal_read;
	int write_signal = p_tile_data->signal_write;

	int new_signal = p_ctrl->n_signal++;
	p_tile_data->signal_write = new_signal;

#ifdef CTRL_DEBUG
	printf("Ctrl_XPhi_EvalTaskMoveTo offload_trans in p_tile: %p p_data: %p length: %i\n", p_tile, p_data, data_size); fflush(stdout);
	printf("Ctrl_XPhi_EvalTaskMoveTo signal_read = %i, signal_write = %i, wait = (read %i, write %i) signal = %i", p_tile_data->signal_read, p_tile_data->signal_write, read_signal, write_signal, new_signal); fflush(stdout);
	printf("\n"); fflush(stdout);
#endif
	#pragma offload_transfer target(mic) in(p_data: length(data_size) align(64) alloc_if(false) free_if(false)) stream(stream) wait(read_signal,write_signal) signal(new_signal)
	
	
	//TDDO @eduardo align(64) cuando podamos asegurar que estan alineados
	if(p_ctrl->policy == CTRL_POLICY_SYNC) {
		printf("SYNC2\n");
		//#pragma offload_wait target(mic:n_xphi) wait(new_signal)
		#pragma offload_wait target(mic) stream(0)
	}

	if (!from_attach && p_ctrl->policy == CTRL_POLICY_SYNC) {
		sem_post(&(p_ctrl->sem_sync));
	}
}

/* 
 * Evaluation of comunication of tile from device to host
 * Params:
 *      cntrl: Controller in charge of task.
 *      run: task that must be evaluated.
 */
void Ctrl_XPhi_EvalTaskMoveFrom(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task, bool from_dettach) {

	HitTile *p_tile = (HitTile *)(p_task->p_arguments);
	Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)(p_tile->ext);

	int n_xphi = p_ctrl->n_xphi;
	char* p_data = (char*) p_tile->data;
	size_t data_size = p_tile->acumCard * p_tile->baseExtent;

	_Offload_stream stream = p_tile_data->stream;

	int write_signal = p_tile_data->signal_write;

	int new_signal = p_ctrl->n_signal++;
	p_tile_data->signal_read = new_signal;

#ifdef CTRL_DEBUG
	printf("Ctrl_XPhi_EvalTaskMoveFrom offload_trans nocopy p_tile: %p p_data: %p length: %i\n", p_tile, p_data, data_size); fflush(stdout);
	printf("Ctrl_XPhi_EvalTaskMoveFrom signal_read = %i, signal_write = %i, wait = %i signal = %i", p_tile_data->signal_read, p_tile_data->signal_write, write_signal, new_signal); fflush(stdout);
	printf("\n"); fflush(stdout);
#endif
	#pragma offload_transfer target(mic) out(p_data: length(data_size) align(64) alloc_if(false) free_if(false)) stream(stream) wait(write_signal) signal(new_signal)

	//TDDO @eduardo align(64) cuando podamos asegurar que estan alineados
	if(p_ctrl->policy == CTRL_POLICY_SYNC) {
		#pragma offload_wait target(mic:n_xphi) wait(new_signal)
	}

	if (!from_dettach && p_ctrl->policy == CTRL_POLICY_SYNC) {
		sem_post(&(p_ctrl->sem_sync));
	}
}


/* 
 * Evaluation of synchronization with tile.
 * Params:
 *      cntrl: Controller in charge of task.
 *      run: task that must be evaluated.
 */
void Ctrl_XPhi_EvalTaskWaitTile(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_arguments);
	Ctrl_XPhi_Tile *p_tile_data = (Ctrl_XPhi_Tile *)(p_tile->ext);

	int n_xphi = p_ctrl->n_xphi;

	int read_signal = p_tile_data->signal_read;

	#pragma offload_wait target(mic:n_xphi)  wait(read_signal)

	sem_post(&(p_ctrl->sem_sync));
}

// void Ctrl_XPhi_EvalTaskDestroyInternal(Ctrl_XPhi *p_ctrl, Ctrl_Task *p_task) {

// }
