/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include <stdio.h>
#include <math.h>
#include <omp.h>
#include <stdbool.h>

#include "Ctrl.h"

#define SEED 6834723

#define STR_SIZE 256

#define EXPAND_RATE 2 // add one iteration will extend the pyramid base by 2 per each borderline

/* maximum power density possible (say 300W for a 10mm x 10mm chip)	*/
#define MAX_PD	(3.0e6)
/* required precision in degrees	*/
#define PRECISION	0.001
#define SPEC_HEAT_SI 1.75e6
#define K_SI 100
/* capacitance fitting factor	*/
#define FACTOR_CHIP	0.5

#define IN_RANGE(x, min, max)   ((x)>=(min) && (x)<=(max))
#define MIN(a, b) ((a)<=(b) ? (a) : (b))

/* chip parameters	*/
float t_chip = 0.0005;
float chip_height = 0.016;
float chip_width = 0.016;
/* ambient temperature, assuming no package at all	*/
float amb_temp = 80.0;

double main_clock;
double exec_clock;

Ctrl_NewType( float );

//CTRL_KERNEL_CHAR(Hotspot, MANUAL, LOCAL_SIZE_0, LOCAL_SIZE_1);

CTRL_KERNEL(Hotspot, OPENCL_GPU, int iteration, KHitTile_float power, KHitTile_float temp_src, KHitTile_float temp_dst,
		int grid_cols, int grid_rows, int border_cols, int border_rows, float Cap, float Rx,
		float Ry, float Rz, float step, float time_elapsed, 
{
	__local float temp_on_cuda[LOCAL_SIZE_0][LOCAL_SIZE_1];
	__local float power_on_cuda[LOCAL_SIZE_0][LOCAL_SIZE_1];
	__local float temp_t[LOCAL_SIZE_0][LOCAL_SIZE_1]; // saving temparary temperature result

	float amb_temp = 80.0;
	float step_div_Cap;
	float Rx_1;
	float Ry_1;
	float Rz_1;

	int bx = get_group_id(0);
	int by = get_group_id(1);

	int tx = get_local_id(0);
	int ty = get_local_id(1);

	step_div_Cap = step / Cap;

	Rx_1 = 1 / Rx;
	Ry_1 = 1 / Ry;
	Rz_1 = 1 / Rz;

	// each block finally computes result for a small block
	// after N iterations.
	// it is the non-overlapping small blocks that cover
	// all the input data

	// calculate the small block size
	int small_block_rows = LOCAL_SIZE_0 - iteration * 2; //EXPAND_RATE
	int small_block_cols = LOCAL_SIZE_1 - iteration * 2; //EXPAND_RATE

	// calculate the boundary for the block according to
	// the boundary of its small block
	int blkY = small_block_rows * by - border_rows;
	int blkX = small_block_cols * bx - border_cols;
	int blkYmax = blkY + LOCAL_SIZE_0 - 1;
	int blkXmax = blkX + LOCAL_SIZE_1 - 1;

	// calculate the global thread coordination
	int yidx = blkY + ty;
	int xidx = blkX + tx;

	// load data if it is within the valid input range
	int loadYidx = yidx;
	int loadXidx = xidx;
	int index = grid_cols * loadYidx + loadXidx;

	if (IN_RANGE(loadYidx, 0, grid_rows - 1) && IN_RANGE(loadXidx, 0, grid_cols - 1)) {
		temp_on_cuda[ty][tx] = hit(temp_src, index); // Load the temperature data from global memory to shared memory
		power_on_cuda[ty][tx] = hit(power, index);   // Load the power data from global memory to shared memory
	}
	barrier(CLK_LOCAL_MEM_FENCE);

	// effective range within this block that falls within
	// the valid range of the input data
	// used to rule out computation outside the boundary.
	int validYmin = (blkY < 0) ? -blkY : 0;
	int validYmax = (blkYmax > grid_rows - 1) ? LOCAL_SIZE_1 - 1 - (blkYmax - grid_rows + 1) : LOCAL_SIZE_1 - 1;
	int validXmin = (blkX < 0) ? -blkX : 0;
	int validXmax = (blkXmax > grid_cols - 1) ? LOCAL_SIZE_0 - 1 - (blkXmax - grid_cols + 1) : LOCAL_SIZE_0 - 1;

	int N = ty - 1;
	int S = ty + 1;
	int W = tx - 1;
	int E = tx + 1;

	N = (N < validYmin) ? validYmin : N;
	S = (S > validYmax) ? validYmax : S;
	W = (W < validXmin) ? validXmin : W;
	E = (E > validXmax) ? validXmax : E;

	bool computed;
	for (int i = 0; i < iteration; i++) {
		computed = false;
		if (IN_RANGE(tx, i + 1, LOCAL_SIZE_1 - i - 2) &&
			IN_RANGE(ty, i + 1, LOCAL_SIZE_0 - i - 2) &&
			IN_RANGE(tx, validXmin, validXmax) &&
			IN_RANGE(ty, validYmin, validYmax))
		{
			computed = true;
			temp_t[ty][tx] = temp_on_cuda[ty][tx] + step_div_Cap * (power_on_cuda[ty][tx] +
								(temp_on_cuda[S][tx] + temp_on_cuda[N][tx] - 2.0 * temp_on_cuda[ty][tx]) * Ry_1 +
								(temp_on_cuda[ty][E] + temp_on_cuda[ty][W] - 2.0 * temp_on_cuda[ty][tx]) * Rx_1 +
								(amb_temp - temp_on_cuda[ty][tx]) * Rz_1);
		}
		barrier(CLK_LOCAL_MEM_FENCE);
		if (i == iteration - 1){
			break;
		}
		if (computed) { //Assign the computation range
			temp_on_cuda[ty][tx] = temp_t[ty][tx];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	// update the global memory
	// after the last iteration, only threads coordinated within the
	// small block perform the calculation and switch on ``computed''
	if (computed) {
		hit(temp_dst, index) = temp_t[ty][tx];
	}
});


CTRL_HOST_TASK( Read_Input, HitTile_float matrix, char *file_name, 
{
	FILE *fp;
	char str[STR_SIZE];
	float val;

	if( (fp  = fopen(file_name, "r" )) ==0 ) {
		fprintf(stderr, "The file was not opened" );
		exit(EXIT_FAILURE);
	}

	for (int i = 0; i < hit_tileDimCard( matrix, 0 ); i++) {
		for (int j = 0; j < hit_tileDimCard( matrix, 1 ); j++) {
			/* This check was not in the original hotspot program */
			if (fgets(str, STR_SIZE, fp) == NULL) {
				fprintf(stderr, "Error reading file\n");
				exit(EXIT_FAILURE);
			}	
			if (feof(fp)) {
				fprintf(stderr, "not enough lines in file");
				/* This call was not in the original hotspot program */
				exit(EXIT_FAILURE);
			}
			if ((sscanf(str, "%f", &val) != 1)) {
				fprintf(stderr, "invalid file format");
				/* This call was not in the original hotspot program */
				exit(EXIT_FAILURE);
			}
			hit(matrix, i, j) = val;
		}
	}

	fclose(fp);
});

CTRL_HOST_TASK( Write_Output, HitTile_float matrix, char * file_name,
{
	int index = 0;
	FILE *fp;
	char str[STR_SIZE];

	if( (fp = fopen(file_name, "w" )) == 0 ) {
		fprintf(stderr, "The file was not opened\n" );
		exit(EXIT_FAILURE);
	}

	for (int i = 0; i < hit_tileDimCard( matrix, 0 ); i++) {
		for (int j = 0; j < hit_tileDimCard( matrix, 1 ); j++) {
			sprintf(str, "%d\t%g\n", index, hit(matrix, i, j));
			fputs(str,fp);
			index++;
		}
	}

	fclose(fp);
});

CTRL_HOST_TASK( Write_Matrix, HitTile_float matrix, char * file_name,
{
	printf("\n MATRIX %s \n", file_name);
	for (int i = 0; i < hit_tileDimCard( matrix, 0 ); i++) {
		for (int j = 0; j < hit_tileDimCard( matrix, 1 ); j++) {
			printf("%f ", hit(matrix, i, j));
		} printf("\n");
	}

	printf("\n\n");
});


CTRL_HOST_TASK( Init_Tiles, HitTile_float matrix_temp, HitTile_float matrix_power,
{
	srand(SEED);
	for (int i = 0; i < hit_tileDimCard( matrix_temp, 0 ); i++) {
		for (int j = 0; j < hit_tileDimCard( matrix_temp, 1 ); j++) {
			hit(matrix_temp, i, j) = (-1 + (2 * (((float)rand())/RAND_MAX)));
		}
	}
	for (int i = 0; i < hit_tileDimCard( matrix_power, 0 ); i++) {
		for (int j = 0; j < hit_tileDimCard( matrix_power, 1 ); j++) {
			hit(matrix_power, i, j) = (-1 + (2 * (((float)rand())/RAND_MAX)));
		}
	}
});

CTRL_HOST_TASK( Host_Compute, HitTile_float matrix_dst, HitTile_float matrix_src,
{
	memcpy(matrix_dst.data, matrix_src.data, sizeof(float)*hit_tileCard( matrix_src ) );
});

CTRL_HOST_TASK( Norm_Calc, HitTile_float matrix,
{
	double resultado = 0;
	double suma = 0;
	
	for (int i = 0; i < hit_tileDimCard( matrix, 0 ); i++) {
		for (int j = 0; j < hit_tileDimCard( matrix, 1 ); j++) {
			suma += pow(hit(matrix, i, j), 2);
		}
	}
	resultado = sqrt(suma);
	#ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("\"sum\":\"%lf\",\"result\":\"%lf\",", suma, resultado);
		fflush(stdout);
	#else
		printf("\n ----------------------- NORM ----------------------- \n\n");
		fflush(stdout);
		printf(" Sum: %lf \n", suma);
		fflush(stdout);
		printf(" Result: %lf \n", resultado);
		fflush(stdout);
		printf("\n ---------------------------------------------------- \n");
		fflush(stdout);
	#endif
});

CTRL_KERNEL_PROTO( Hotspot,
    1, OPENCL_GPU, 14, 
	INVAL, int, iteration,
	IN, HitTile_float, power,
	IN, HitTile_float, temp_src,
	OUT, HitTile_float, temp_dst, 
	INVAL, int, grid_cols,
	INVAL, int, grid_rows,
	INVAL, int, border_cols,
	INVAL, int, border_rows,
	INVAL, float, Cap,
	INVAL, float, Rx,
	INVAL, float, Ry,
	INVAL, float, Rz,
	INVAL, float, step,
	INVAL, float, time_elapsed
);

CTRL_HOST_TASK_PROTO( Read_Input, 2, 
	OUT, HitTile_float, matrix, 
	INVAL, char *, file_name
);


CTRL_HOST_TASK_PROTO( Write_Output, 2, 
	INVAL, HitTile_float, matrix, 
	INVAL, char *, file_name
);

CTRL_HOST_TASK_PROTO( Write_Matrix, 2, 
	INVAL, HitTile_float, matrix, 
	INVAL, char *, file_name
);

CTRL_HOST_TASK_PROTO( Init_Tiles, 2, 
	OUT, HitTile_float, matrix_temp, 
	OUT, HitTile_float, matrix_power
);

CTRL_HOST_TASK_PROTO( Host_Compute, 2,
	INVAL, HitTile_float, matrix_dst,
	IN, HitTile_float, matrix_src 
);

CTRL_HOST_TASK_PROTO( Norm_Calc, 1,
	INVAL, HitTile_float, matrix
);

int main(int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	if (argc != 8) {
		fprintf(stderr, "Usage: %s <grid_rows/grid_cols> <pyramid_height> <sim_time> <device> <platform> <policy>\n", argv[0]);
		fprintf(stderr, "\t<grid_rows/grid_cols> - number of rows/cols in the grid (positive integer)\n");
		fprintf(stderr, "\t<pyramid_height> - pyramid heigh(positive integer)\n");
		fprintf(stderr, "\t<sim_time> - number of iterations\n");
		fprintf(stderr, "\t<iters_per_copy> - nº of iter between each copy back\n");
		fprintf(stderr, "\t<device> - GPU index\n");
		fprintf(stderr, "\t<platform> - GPU index\n");
		fprintf(stderr, "\t<policy> - 0 - Sync, 1 - Async\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	int grid_rows = atoi(argv[1]);
	int grid_cols = atoi(argv[1]);
	int pyramid_height = atoi(argv[2]);
	int total_iterations = atoi(argv[3]);
	int iters_per_copy = atoi(argv[4]);
	int DEVICE = atoi(argv[5]);
	int PLATFORM = atoi(argv[6]);
	int POLICY = atoi(argv[7]);

	int borderCols = (pyramid_height)*EXPAND_RATE / 2;
	int borderRows = (pyramid_height)*EXPAND_RATE / 2;
	int smallBlockCol = LOCAL_SIZE_0 - (pyramid_height)*EXPAND_RATE;
	int smallBlockRow = LOCAL_SIZE_1 - (pyramid_height)*EXPAND_RATE;
	int blockCols = grid_cols / smallBlockCol + ((grid_cols % smallBlockCol == 0) ? 0 : 1);
	int blockRows = grid_rows / smallBlockRow + ((grid_rows % smallBlockRow == 0) ? 0 : 1);

	float grid_height = chip_height / grid_rows;
	float grid_width = chip_width / grid_cols;

	float Cap = FACTOR_CHIP * SPEC_HEAT_SI * t_chip * grid_width * grid_height;
	float Rx = grid_width / (2.0 * K_SI * t_chip * grid_height);
	float Ry = grid_height / (2.0 * K_SI * t_chip * grid_width);
	float Rz = t_chip / (K_SI * grid_height * grid_width);

	float max_slope = MAX_PD / (FACTOR_CHIP * t_chip * SPEC_HEAT_SI);
	float step = PRECISION / max_slope;
	float time_elapsed = 0.001;

	#ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("{\"size\":%ld,\"height\":%d,\"n_iter\":%d,\"iter_per_copy\":%d,\"arch\":\"opencl\",\"device\":%d, \"version\":\"ctrl\",", 
				grid_rows, pyramid_height, total_iterations, iters_per_copy, DEVICE);
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\"policy\":\"async\",");
		#else  // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\"policy\":\"sync\",");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		fflush(stdout);
	#endif

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
		#ifndef _CTRL_EXAMPLES_EXP_MODE_
			printf("\n ----------------------- ARGS ----------------------- \n");
			printf("\n SIZE (SIZE x SIZE): %ld, %ld, %ld", grid_rows * grid_cols, grid_rows, grid_cols);
			printf("\n PYRAMID HEIGHT: %d", pyramid_height);
			printf("\n N_ITER: %d", total_iterations);
			printf("\n ITERS_PER_COPY: %d", iters_per_copy);
			printf("\n DEVICE: %d", DEVICE);
			printf("\n PLATFORM: %d", PLATFORM);
			#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY ASYNC");
			#else  // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY SYNC");
			#endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\n\n ---------------------------------------------------- \n");
			fflush(stdout);
		#endif // _CTRL_EXAMPLES_EXP_MODE_
    #endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

	Ctrl_Thread threads;
	#ifndef _CTRL_EXAMPLES_EXP_MODE_
		printf("Threads: %d, %d\n", LOCAL_SIZE_0 * blockRows, LOCAL_SIZE_1 * blockCols);
		fflush(stdout);
	#endif

	Ctrl_ThreadInit(threads, LOCAL_SIZE_0 * blockRows, LOCAL_SIZE_1 * blockCols);

	Ctrl_Thread group;
	Ctrl_ThreadInit(group, LOCAL_SIZE_0, LOCAL_SIZE_1);

	__ctrl_block__(1)
	{
		
		PCtrl ctrl = Ctrl_Create(CTRL_TYPE_OPENCL_GPU, POLICY, DEVICE, PLATFORM);

		HitTile_float MatrixTemp[2], MatrixPower;
		MatrixTemp[0] = Ctrl_Alloc(ctrl, float, hitNewShapeSize(grid_rows, grid_cols) );
		MatrixTemp[1] = Ctrl_Alloc(ctrl, float, hitNewShapeSize(grid_rows, grid_cols) );
		MatrixPower = Ctrl_Alloc(ctrl, float, hitNewShapeSize(grid_rows, grid_cols));
		HitTile_float MatrixCopy = hitNewTile(float, hitNewShapeSize(grid_rows, grid_cols));

		Ctrl_Launch(ctrl, Hotspot, threads, group, pyramid_height, MatrixPower, MatrixTemp[0], MatrixTemp[0], grid_cols,
					grid_rows, borderCols, borderRows, Cap, Rx, Ry, Rz, step, time_elapsed);


		Ctrl_HostTask(ctrl, Init_Tiles, MatrixTemp[0], MatrixPower);

		Ctrl_GlobalSync(ctrl);
		exec_clock = omp_get_wtime();

		Ctrl_MoveTo(ctrl, MatrixTemp[0], MatrixPower);

		int real_iter = 1;
		int src = 1;
		int dst = 0;

		for (int i = 0; i < total_iterations; i += pyramid_height) {
			int temp = src;
			src = dst;
			dst = temp;

			int aux_iterations = MIN(pyramid_height, total_iterations - i);
			Ctrl_Launch(ctrl, Hotspot, threads, group, aux_iterations, MatrixPower, MatrixTemp[src], MatrixTemp[dst], grid_cols,
					grid_rows, borderCols, borderRows, Cap, Rx, Ry, Rz, step, time_elapsed);

			if ((real_iter % iters_per_copy) == 0) {
				Ctrl_MoveFrom(ctrl, MatrixTemp[dst]);
				Ctrl_HostTask(ctrl, Host_Compute, MatrixCopy, MatrixTemp[dst]);
			}
			real_iter++;
		}

		Ctrl_MoveFrom(ctrl, MatrixTemp[dst]);
		Ctrl_HostTask(ctrl, Host_Compute, MatrixCopy, MatrixTemp[dst]);

		Ctrl_GlobalSync(ctrl);
		exec_clock = omp_get_wtime() - exec_clock;

		Ctrl_HostTask(ctrl, Norm_Calc, MatrixCopy);

		Ctrl_GlobalSync(ctrl);
		Ctrl_Free(ctrl, MatrixTemp[0], MatrixTemp[1], MatrixPower);
		hit_tileFree(MatrixCopy);
		Ctrl_Destroy(ctrl);
	}

	main_clock = omp_get_wtime() - main_clock;

	#ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("\"main_clock\":%lf,\"exec_clock\":%lf}\n", main_clock, exec_clock);
		fflush(stdout);
	#else
		printf("\n ----------------------- TIME ----------------------- \n\n");
		printf(" Clock main: %lf\n", main_clock);
		printf(" Clock exec: %lf\n", exec_clock);
		printf("\n ---------------------------------------------------- \n");
	#endif

	return 0;
}
