#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#include "Ctrl.h"

#define SEED 6834723

#define BLOCK_SIZE 16

Ctrl_NewType( float );

double main_clock;
double exec_clock;


CTRL_KERNEL(Mult, CUDA, KHitTile_float matrix_result, KHitTile_float matrix_a, KHitTile_float matrix_b, int n,
{
	__shared__ float tile_a[BLOCKSIZE_0][BLOCKSIZE_1];
	__shared__ float tile_b[BLOCKSIZE_0][BLOCKSIZE_1];

	float tmp = 0.0;
	int idx;
    
	for (int sub = 0; sub < gridDim.x; ++sub) {
		idx = thread_id.x * n + sub * BLOCKSIZE_0 + threadIdx.x;
		if (idx >= n * n) {
			tile_a[threadIdx.y][threadIdx.x] = 0;
		} else {
			tile_a[threadIdx.y][threadIdx.x] = hit(matrix_a, idx);
		}
		idx = (sub * BLOCKSIZE_1 + threadIdx.y) * n + thread_id.y;
		if (idx >= n * n) {
			tile_b[threadIdx.y][threadIdx.x] = 0;
		} else {
			tile_b[threadIdx.y][threadIdx.x] = hit(matrix_b, idx);
		}
		__syncthreads();
	
		for (int k = 0; k < BLOCKSIZE_0; ++k) {
			tmp += tile_a[threadIdx.y][k] * tile_b[k][threadIdx.x];
		}
		__syncthreads();
	}
	
	if (thread_id.x < n && thread_id.y < n) {
		hit(matrix_result, thread_id.x, thread_id.y) = tmp;
	}
});

CTRL_HOST_TASK( Init_Tiles, HitTile_float matrix_a, HitTile_float matrix_b, HitTile_float matrix_c,
{
	srand(SEED);
	for (int i = 0; i < hit_tileDimCard( matrix_a, 0 ); i++) {
		for (int j = 0; j < hit_tileDimCard( matrix_a, 1 ); j++) {
			hit(matrix_a, i, j) = (-1 + (2 * (((float)rand())/RAND_MAX)));
			hit(matrix_b, i, j) = (-1 + (2 * (((float)rand())/RAND_MAX)));
			hit(matrix_c, i, j) = 0;
		}
	}
});

CTRL_HOST_TASK( Host_Compute, int ITER, double *p_sum, double *p_res, HitTile_float matrix, HitTile_float matrix_res,
{
	double minimum = hit(matrix, 0, 0);
	double maximum = hit(matrix, 0, 0);

	for (int j = 0; j < hit_tileDimCard( matrix, 0 ); j++) {
		for (int k = 0; k < hit_tileDimCard( matrix, 1 ); k++) {
			if (minimum > hit(matrix, j, k)){
				minimum = hit(matrix, j, k);
			}
			if (maximum < hit(matrix, j, k)){
				maximum = hit(matrix, j, k);
			}
		}
	}

	for (int j = 0; j < hit_tileDimCard( matrix, 0 ); j++) {
		for (int k = 0; k < hit_tileDimCard( matrix, 1 ); k++) {
			hit(matrix, j, k) = hit(matrix, j, k) - minimum;
			hit(matrix, j, k) = hit(matrix, j, k) / maximum; 
		}
	}

	p_sum[ITER] = 0;
	for (int j = 0; j < hit_tileDimCard( matrix, 0 ); j++) {
		for (int k = 0; k < hit_tileDimCard( matrix, 1 ); k++) {
			p_sum[ITER] += pow(hit(matrix, j, k), 2);
		}
	}
	p_res[ITER] = sqrt(p_sum[ITER]);

	for (int j = 0; j < hit_tileDimCard( matrix, 0 ); j++) {
		for (int k = 0; k < hit_tileDimCard( matrix, 1 ); k++) {
			hit(matrix, j, k) = hit(matrix, j, k) / p_res[ITER];
		}
	}				
	memcpy(matrix_res.data, matrix.data, sizeof(float) * hit_tileDimCard( matrix, 0 ) * hit_tileDimCard( matrix, 1 ));
});

CTRL_KERNEL_PROTO( Mult,
	1, CUDA, 4, 
	OUT, HitTile_float, matrix_result, 
	IN, HitTile_float, matrix_a, 
	IN, HitTile_float, matrix_b,
	INVAL, int, n
);

CTRL_HOST_TASK_PROTO( Init_Tiles, 3,
	OUT, HitTile_float, matrix_a,
	OUT, HitTile_float, matrix_b, 
	OUT, HitTile_float, matrix_c
);

CTRL_HOST_TASK_PROTO( Host_Compute, 5,
	INVAL, int, ITER,
	INVAL, double *, p_sum,
	INVAL, double *, p_res, 
	IN, HitTile_float, matrix,
	INVAL, HitTile_float, matrix_res
);


int main(int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	if (argc != 5) {
		fprintf(stderr, "\nUsage: %s <size> <n_iter> <device> <policy>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	int SIZE = atoi(argv[1]);
	int N_ITER = atoi(argv[2]);
	int DEVICE = atoi(argv[3]);
	Ctrl_Policy POLICY = (Ctrl_Policy)atoi(argv[4]);

	#ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("{\"size\":%d,\"n_iter\":%d,\"arch\":\"cuda\",\"device\":%d, \"version\":\"ctrl\",", 
				SIZE, N_ITER, DEVICE);
		#ifdef _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
			printf("\"policy\":\"async\",");
		#else  // _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
			printf("\"policy\":\"sync\",");
		#endif // _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
		fflush(stdout);
	#endif

	#ifdef _CTRL_EXAMPLES_CUDA_INFO_
		#ifndef _CTRL_EXAMPLES_EXP_MODE_
			printf("\n ----------------------- ARGS ----------------------- \n");
			printf("\n SIZE: %d", SIZE);
			printf("\n N_ITER: %d", N_ITER);
			printf("\n DEVICE: %d", DEVICE);
			if(POLICY == CTRL_POLICY_ASYNC)
				printf("\n POLICY ASYNC");
			else if(POLICY == CTRL_POLICY_SYNC)
				printf("\n POLICY SYNC");
			printf("\n\n ---------------------------------------------------- \n");
			fflush(stdout);
		#endif // _CTRL_EXAMPLES_EXP_MODE_
	#endif // _CTRL_EXAMPLES_CUDA_INFO_

	double *p_res = (double *)malloc(N_ITER * sizeof(double));
	double *p_sum = (double *)malloc(N_ITER * sizeof(double));

	Ctrl_Thread threads;
	Ctrl_ThreadInit(threads, SIZE, SIZE);

	Ctrl_Thread block;
	Ctrl_ThreadInit(block, BLOCKSIZE_1, BLOCKSIZE_1);
	__ctrl_block__(1)
	{

		PCtrl ctrl = Ctrl_Create(CTRL_TYPE_CUDA, POLICY, DEVICE);

		HitTile_float matrix_a = Ctrl_Alloc(ctrl, float, hitNewShapeSize(SIZE, SIZE) );
		HitTile_float matrix_b = Ctrl_Alloc(ctrl, float, hitNewShapeSize(SIZE, SIZE) );
		HitTile_float matrix_c = Ctrl_Alloc(ctrl, float, hitNewShapeSize(SIZE, SIZE) );

		HitTile_float matrix_tmp = hitNewTile(float, hitNewShapeSize(SIZE, SIZE));

		Ctrl_HostTask(ctrl, Init_Tiles, matrix_a, matrix_b, matrix_c);

		Ctrl_GlobalSync(ctrl);
		exec_clock = omp_get_wtime();

		// Ctrl_MoveTo(ctrl, matrix_a, matrix_b, matrix_c);

        for (int i = 0; i < N_ITER; i++) {
			if ((i % 2) == 0) {
				Ctrl_Launch(ctrl, Mult, threads, block, matrix_c, matrix_a, matrix_b, SIZE);
				// Ctrl_MoveFrom(ctrl, matrix_c);
				Ctrl_HostTask(ctrl, Host_Compute, i, p_sum, p_res, matrix_c, matrix_tmp);
			} else {
				Ctrl_Launch(ctrl, Mult, threads, block, matrix_b, matrix_a, matrix_c, SIZE);
				// Ctrl_MoveFrom(ctrl, matrix_b);
				Ctrl_HostTask(ctrl, Host_Compute, i, p_sum, p_res, matrix_b, matrix_tmp);
			}
        }

		Ctrl_GlobalSync(ctrl);
		exec_clock = omp_get_wtime() - exec_clock;

		#ifdef _CTRL_EXAMPLES_EXP_MODE_
			printf("\"results\":[");
			for (int i = 0; i < N_ITER; i++) {
				printf("{\"iter\":%d,\"sum\":\"%lf\",\"res\":\"%lf\"}", i + 1, p_sum[i], p_res[i]);
				if (i+1 != N_ITER) {
					printf(",");
				}
			}
			printf("],");
			fflush(stdout);
		#else	
			printf("\n ----------------------- NORM ----------------------- \n\n"); fflush(stdout);
			for (int i = 0; i < N_ITER; i++) {
				printf(" iter: %d, sum: %lf, res: %lf\n", i + 1, p_sum[i], p_res[i]); fflush(stdout);
			}
			printf("\n ---------------------------------------------------- \n"); fflush(stdout);
		#endif

		Ctrl_Free(ctrl, matrix_a, matrix_b, matrix_c);
		hit_tileFree(matrix_tmp);
			
		Ctrl_Destroy(ctrl);
	}

	free(p_sum);
	free(p_res);

	main_clock = omp_get_wtime() - main_clock;

	#ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("\"main_clock\":%lf,\"exec_clock\":%lf}\n", main_clock, exec_clock);
		fflush(stdout);
	#else
		printf("\n ----------------------- TIME ----------------------- \n\n");
		printf(" Clock main: %lf\n", main_clock);
		printf(" Clock exec: %lf\n", exec_clock);
		printf("\n ---------------------------------------------------- \n");
	#endif

	return 0;
}
