/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include <math.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
// #include <sys/mman.h>

#include "Sobel_YUV_FPGA_Ctrl.h"

#define ERR_Y -1
#define ERR_U -2
#define ERR_V -3

#define SEED 6834723

/*Tiempos*/
double main_clock;
double exec_clock;


#define THRESH_HOLD_Y 20
#define THRESH_HOLD_U 1
#define THRESH_HOLD_V 1

#define N_IMG 3
#define IMG_Y 0
#define IMG_U 1
#define IMG_V 2

void Save_Frame(BYTE *Output_Img[N_IMG], FILE *File_writer, size_t sizes[N_IMG]) {
    for (int i = 0; i < N_IMG; i++) {
        fwrite(Output_Img[i], sizeof(BYTE), sizes[i], File_writer);
    }
}


CTRL_HOST_TASK(Load_Frame, HitTile_BYTE Image_Y, HitTile_BYTE Image_U,
        HitTile_BYTE Image_V, FILE *File_reader, size_t *sizes,
{
    if(!(fread(&(hit(Image_Y, 0)), sizeof(BYTE), sizes[IMG_Y], File_reader))) {
        printf("Cannot read frame Y\n");
        exit(ERR_Y);
    }
    
    if(!(fread(&(hit(Image_U, 0)), sizeof(BYTE), sizes[IMG_U], File_reader))) {
        printf("Cannot read frame U\n");
        exit(ERR_U);
    }
    
    if(!(fread(&(hit(Image_V, 0)), sizeof(BYTE), sizes[IMG_V], File_reader))) {
        printf("Cannot read frame V\n");
        exit(ERR_V);
    }
});


CTRL_HOST_TASK_PROTO(Load_Frame, 5, 
        OUT, HitTile_BYTE, Image_Y,
        OUT, HitTile_BYTE, Image_U,
        OUT, HitTile_BYTE, Image_V,
        INVAL, FILE *, File_reader,
        INVAL, size_t *, sizes
);

CTRL_HOST_TASK(Put_Frame, HitTile_BYTE Image_Y, HitTile_BYTE Image_U,
        HitTile_BYTE Image_V, BYTE **buffer_write, size_t *sizes,
{
    memcpy(buffer_write[IMG_Y], &(hit(Image_Y, 0)), sizeof(BYTE) * sizes[IMG_Y]);
    memcpy(buffer_write[IMG_U], &(hit(Image_U, 0)), sizeof(BYTE) * sizes[IMG_U]);
    memcpy(buffer_write[IMG_V], &(hit(Image_V, 0)), sizeof(BYTE) * sizes[IMG_V]);
});

CTRL_HOST_TASK_PROTO(Put_Frame, 5,
        IN, HitTile_BYTE, Image_Y,
        IN, HitTile_BYTE, Image_U,
        IN, HitTile_BYTE, Image_V,
        INVAL, BYTE **, buffer_write,
        INVAL, size_t *, sizes
);

int main(int argc, char *argv[]) {
    main_clock = omp_get_wtime();

    if (argc != 10) {
        printf("Usage: %s <width> <height> <num_frames> <input_yuv_file> "
               "<output_yuv_file> <device> <platform>\n",
               argv[0]);
        exit(EXIT_FAILURE);
    }

    int Width[3];
    Width[0] = atoi(argv[1]);
    Width[1] = Width[2] = Width[0] / 2; 
    
    int Height[3];
    Height[0] = atoi(argv[2]);
    Height[1] = Height[2] = Height[0] / 2;
    
    int Num_Frames = atoi(argv[3]);
    
    char *Input_Filename = argv[4];
    char *Output_Filename = argv[5];
    int DEVICE = atoi(argv[6]);
    int PLATFORM = atoi(argv[7]);
    Ctrl_Policy POLICY = atoi(argv[8]);
    int EXEC_MODE = atoi(argv[9]);

    int Frame_num = 0; // loop variable

    int thresh_hold[N_IMG] = {
            THRESH_HOLD_Y,
            THRESH_HOLD_U,
            THRESH_HOLD_V
    };

    size_t *sizes = (size_t *)malloc(sizeof(size_t) * N_IMG);
    sizes[IMG_Y] = (size_t)(Width[IMG_Y] * Height[IMG_Y]);
    sizes[IMG_U] = (size_t)(Width[IMG_U] * Height[IMG_U]);
    sizes[IMG_V] = (size_t)(Width[IMG_V] * Height[IMG_V]);

    BYTE ***buffer_write;
    
    buffer_write = (BYTE ***)malloc(sizeof(BYTE **) * Num_Frames);

    for (int i = 0; i < Num_Frames; i++) {
        buffer_write[i] = (BYTE **)malloc(sizeof(BYTE *) * N_IMG);

        for (int j = 0; j < N_IMG; j++) {
            buffer_write[i][j] = (BYTE*)malloc(sizes[j] * sizeof(BYTE));
        }
    }

    // File pointer for reading and writting
    FILE *File_writer, *File_reader;


    #ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("{\"width\":%d,\"height\":%d,\"num_frames\":%d,\"version\":\"ctrl\",\"arch\":\"opencl\",\"device\":%d,\"platform\":%d,", 
				Width[0], Height[0], Num_Frames, DEVICE, PLATFORM);
		if(POLICY == CTRL_POLICY_ASYNC)
			printf("\"policy\":\"async\",");
		else
			printf("\"policy\":\"sync\",");
		fflush(stdout);
	#endif

    #ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
		#ifndef _CTRL_EXAMPLES_EXP_MODE_
            printf("\n ----------------------- ARGS ----------------------- \n");
            printf("\n WIDTH: %d", Width[0]);
            printf("\n HEIGHT: %d", Height[0]);
            printf("\n NUM_FRAMES: %d", Num_Frames);
            printf("\n DEVICE: %d", DEVICE);
            printf("\n PLATFORM: %d", PLATFORM);
            #ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
                printf("\n POLICY ASYNC");
            #else  // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
                printf("\n POLICY SYNC");
            #endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
            printf("\n\n ---------------------------------------------------- \n");
            fflush(stdout);
		#endif // _CTRL_EXAMPLES_EXP_MODE_
    #endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

    Ctrl_Thread threads[N_IMG];
    for (int i = 0; i < N_IMG; i++) {
        Ctrl_ThreadInit(threads[i], Height[i], Width[i]);
    }
    
    if (!(File_reader = fopen(Input_Filename, "rb"))) {
        printf("\nError in opening input file: %s\n", Input_Filename);
        exit(EXIT_FAILURE);
    }
    if (!(File_writer = fopen(Output_Filename, "wb+"))) {
        printf("\nError in opening output file: %s\n", Output_Filename);
        exit(EXIT_FAILURE);
    }
    
    Ctrl_Thread group;
	Ctrl_ThreadInit(group, 1);
    __ctrl_block__(1)
    {
        PCtrl ctrl = Ctrl_Create(CTRL_TYPE_FPGA, POLICY, DEVICE, PLATFORM, EXEC_MODE);

        HitTile_BYTE Input_Img[N_IMG];
        Input_Img[IMG_Y] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_Y], Height[IMG_Y]));
        Input_Img[IMG_U] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_U], Height[IMG_U]));
        Input_Img[IMG_V] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_V], Height[IMG_V]));

        HitTile_BYTE Output_Img[N_IMG];
        Output_Img[IMG_Y] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_Y], Height[IMG_Y]));
        Output_Img[IMG_U] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_U], Height[IMG_U]));
        Output_Img[IMG_V] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_V], Height[IMG_V]));
        
        int iterations[N_IMG];
        iterations[0] = Width[0] * Height[0];
        iterations[1] = Width[1] * Height[1];
        iterations[2] = Width[2] * Height[2];

        Ctrl_Launch(ctrl, Sobel_Operation, threads[0], group, Input_Img[0], Output_Img[0], 
                iterations[0], thresh_hold[0]
        );
        
        Ctrl_GlobalSync(ctrl);
        exec_clock = omp_get_wtime();

        Ctrl_HostTask(ctrl, Load_Frame, Input_Img[IMG_Y], Input_Img[IMG_U], 
                Input_Img[IMG_V], File_reader, sizes
        );
        
        for (Frame_num = 0; Frame_num < Num_Frames; Frame_num++) {

            for (int i = 0; i < N_IMG; i++) {
                Ctrl_Launch(ctrl, Sobel_Operation, threads[i], group, Input_Img[i], Output_Img[i], 
                        iterations[i], thresh_hold[i]
                );
            }
            if (Frame_num + 1 < Num_Frames) {
                Ctrl_HostTask(ctrl, Load_Frame, Input_Img[IMG_Y], Input_Img[IMG_U], 
                        Input_Img[IMG_V], File_reader, sizes
                );
            }
            Ctrl_HostTask(ctrl, Put_Frame, Output_Img[IMG_Y], Output_Img[IMG_U], 
                    Output_Img[IMG_V], buffer_write[Frame_num], sizes
            );
        }

        Ctrl_GlobalSync(ctrl);
        exec_clock = omp_get_wtime() - exec_clock;

        for (int i = 0; i < Num_Frames; i++) {
            Save_Frame(buffer_write[i], File_writer, sizes);
        }

        Ctrl_Free(ctrl, Input_Img[IMG_Y], Input_Img[IMG_U], Input_Img[IMG_V],
                Output_Img[IMG_Y], Output_Img[IMG_U], Output_Img[IMG_V]
        );

        Ctrl_Destroy(ctrl);
    }

    free(sizes);
    fclose(File_reader);
    fclose(File_writer);

    for (int i = 0; i < Num_Frames; i++) {
        for (int j = 0; j < N_IMG; j++) {
            free(buffer_write[i][j]);
        }
        free(buffer_write[i]);
    }
    free(buffer_write);

    main_clock = omp_get_wtime() - main_clock;
    
    #ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("\"main_clock\":%lf,\"exec_clock\":%lf}\n", main_clock, exec_clock);
		fflush(stdout);
	#else
		printf("\n ----------------------- TIME ----------------------- \n\n");
		printf(" Clock main: %lf\n", main_clock);
		printf(" Clock exec: %lf\n", exec_clock);
        printf("\n ---------------------------------------------------- \n");
	#endif

    return 0;
}
